import argparse
import pickle
import matplotlib.pyplot as matplt
import numpy as np
import os
import scipy
import subprocess
import sys

from datetime import datetime
from pathlib import Path

import gnssrefl.phase_functions as qp
import gnssrefl.gps as g
import gnssrefl.gnssir_v2 as gnssir

from gnssrefl.utils import str2bool, read_files_in_dir

xdir = os.environ['REFL_CODE']


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("station", help="station", type=str)
    parser.add_argument("year", help="year", type=int)
    parser.add_argument("-year_end", default=None, help="year_end", type=int)
    parser.add_argument("-fr", help="frequency", type=int)
    parser.add_argument("-plt", default=None, type=str, help="boolean for plotting to screen")
    parser.add_argument("-screenstats", default=None, type=str, help="boolean for plotting statistics to screen")
    parser.add_argument("-min_req_pts_track", default=None, type=int, help="min number of points for a track to be kept. Default is 50")
    parser.add_argument("-polyorder", default=None, type=int, help="override on polynomial order")
    parser.add_argument("-minvalperday", default=None, type=int, help="min number of satellite tracks needed each day. Default is 10")
    parser.add_argument("-snow_filter", default=None, type=str, help="boolean, try to remove snow contaminated points. Default is F")
    parser.add_argument("-subdir", default=None, type=str, help="use non-default subdirectory for output files")
    parser.add_argument("-tmin", default=None, type=float, help="minimum soil texture. Default is 0.05.")
    parser.add_argument("-tmax", default=None, type=float, help="maximum soil texture. Defafult is 0.50.")
    parser.add_argument("-warning_value", default=None, type=float, help="Phase RMS (deg) threshold for bad tracks, default is 5.5 degrees")
    parser.add_argument("-auto_removal", default=None, type=str, help="Whether you want to remove bad tracks automatically, default is False")
    parser.add_argument("-hires_figs", default=None, type=str, help="Whether you want eps instead of png files")
    parser.add_argument("-advanced", default=None, type=str, help="Whether you want to implement advanced veg model (in development)")

    g.print_version_to_screen()

    args = parser.parse_args().__dict__

    boolean_args = ['plt','screenstats','snow_filter','auto_removal','hires_figs','advanced']
    args = str2bool(args, boolean_args)
    # only return a dictionary of arguments that were added from the user - all other defaults will be set in code below
    return {key: value for key, value in args.items() if value is not None}


def vwc(station: str, year: int, year_end: int = None, fr: int = 20, plt: bool = True, screenstats: bool = False, 
        min_req_pts_track: int = 150, polyorder: int = -99, minvalperday: int = 10, 
        snow_filter: bool = False, subdir: str=None, tmin: float=None, tmax: float=None, 
        warning_value : float=5.5, auto_removal : bool=False, hires_figs : bool=False, advanced : bool=False ):
    """
    The goal of this code is to compute volumetric water content (VWC) from GNSS-IR phase estimates. 
    It concatenates previously computed phase results, makes plots for the four geographic quadrants, computes daily 
    average phase files before converting to volumetric water content (VWC). It uses the simple vegetation model
    from Clara Chew's dissertation. For the more advanced vegetation model, we will need a volumteer to convert it from Matlab.
    It is not a difficult port - but it will require care be taken that it is checked carefully.


    Code now allows inputs (minvalperday, tmin, and tmax) to be stored in the gnssir analysis json file.
    To avoid confusion, in the json they are called vwc_minvalperday, vwc_min_soil_texture, and vwc_max_soil_texture.
    These values can also be overwritten on the command line ... 

    Examples
    --------
    vwc p038 2017
        one year for station p038 
    vwc p038 2015 -year_end 2017
        three years of analysis for station p038 
    vwc p038 2015 -year_end 2017 -warning_value 6
        warns you about tracks with phase RMS greater than 6 degrees rms 
    vwc p038 2015 -year_end 2017 -warning_value 6 -auto_removal T 
        makes new list of tracks based on your new warning value

    Parameters
    ----------
    station : str
        4 character ID of the station
    year : int
        full Year
    year_end : int, optional
        last year for analysis
    fr : int, optional
        GNSS frequency. Currently only supports l2c.
        Default is 20 (l2c)
    plt: bool, optional
        Whether to produce plots to the screen.
        Default is True
    min_req_pts_track : int, optional
        how many points needed to keep a satellite track
        default is now set to 150 (was previously 50). This is an issue when a satellite has recently 
        been launched. You don't really have enough information to trust it for several months (and in some cases longer)
        this can now be set in the gnssir_input created analysis json (vwc_min_req_pts_track)
    polyorder : int
        polynomial order used for leveling.  Usually the code picks it but this allows to users to override. 
        Default is -99 which means let the code decide
    minvalperday: integer
        how many phase measurements are needed for each daily measurement
        default is 10
    snow_filter: bool
        whether you want to attempt to remove points contaminated by snow
        default is False
    subdir: str
        subdirectory in $REFL_CODE/Files for plots and text file outputs
    tmin: float
        minimum soil texture value, default 0.05. This can now be set in the gnssir_input json (with vwc_ added)
    tmax: float
        maximum soil texture value, default 0.5. This can now be set in the gnssir_input json (with vwc_ added)
    warning_value : float
         screen warning about bad tracks (phase rms, in degrees).
         default is 5.5 
    auto_removal : bool, optional
         whether to automatically remove tracks that hit your bad track threshold
         default value is false
    hires_figs: bool, optional
         whether to make eps instead of png files
         default value is false
    advanced : bool, optional
         advanced veg model implmentation. Currently in testing

    Returns
    -------

    Daily phase results in a file at $REFL_CODE/Files/<station>/<station>_phase.txt
        with columns: Year DOY Ph Phsig NormA MM DD

    VWC results in a file at $$REFL_CODE/Files/<station>/<station>_vwc.txt
        with columns: FracYr Year DOY  VWC Month Day

    """
    fs =10 # fontsize
    snow_file = None
    colors = 'mrgbcykmrgbcykmrbcykmrgbcykmrgbcykmrbcyk'
    # plotting indices for phase data
    by=[0,1,0,1]; bx=[0,0,1,1]
    # azimuth list, starting point only (i.e. 270 means 270 thru 360)
    azlist = [270, 0, 180,90 ]
    # consistency with old adv vegetation code
    oldquads = [2, 1, 3, 4] # number system from pboh2o, only used for 

    # pick up the parameters used for this code
    minvalperday, tmin, tmax, min_req_pts_track, freq, year_end, subdir, plt, \
            remove_bad_tracks, warning_value,min_norm_amp,sat_legend,circles= \
            qp.set_parameters(station, minvalperday, tmin,tmax, min_req_pts_track, 
                              fr, year, year_end,subdir,plt, auto_removal,warning_value)

    # if you have requested snow filtering
    if snow_filter:
        medf = 0.2 # this is meters , so will only see big snow 
        ReqTracks = 10 # have a pretty small number here
        print('You have chosen the snow filter option')
        snowfileexists,snow_file = qp.make_snow_filter(station, medf, ReqTracks, year, year_end+1)
        matplt.close ('all')# we do not want the plots to come to the screen for the daily average

    # load past VWC analysis  for QC
    avg_exist, avg_date, avg_phase = qp.load_avg_phase(station,freq)

    # pick up all the phase data. unwrapped phase is stored in the results variable
    data_exist, year_sat_phase, doy, hr, phase, azdata, ssat, rh, amp_lsp,amp_ls,ap_rh, results = \
            qp.load_phase_filter_out_snow(station, year, year_end, fr,snow_file)


    if not data_exist:
        print('No data were found. Check your frequency request or station name')
        sys.exit()

    # using unwrapped phase instead of raw phase
    phase = results[17,:]
    
    # pick up the tracks you will use to compute phase.  THis was previously
    # created by vwc_input ...
    tracks = qp.read_apriori_rh(station,freq)
    nr = len(tracks[:,1])

    if (minvalperday > nr ):
        print('The code thinks you are using ', nr, ' satellite tracks but you are requiring', minvalperday, ' for each VWC point.')
        print('Try lowering minvalperday at the command line or in the gnssir analysis json (vwc_minvalperday)')
        sys.exit()
    if (nr < 15 ) and (minvalperday==10):
        print('The code thinks you are using ', nr, ' satellite tracks but that is pretty close to the default (', minvalperday, ')')
        print('This could be problematic. Try lowering minvalperday at the command line or in the gnssir analysis json (vwc_minvalperday)')
        sys.exit()

    atracks = tracks[:, 5]  # min azimuth values
    stracks = tracks[:, 2]  # satellite names

    k = 1
    # define the contents of this variable HERE
    vxyz = np.empty(shape=[0, 15]) 
    # newl = np.vstack((y, t, new_phase, azd, s, rhs, norm_ampLSP,norm_ampLS,h,amp_lsps,amp_lss,qs)).T
    # column, contents of this variable
    # 0 year
    # 1 doy 
    # 2 phase : degrees (unwrapped)
    # 3 azimuth
    # 4 satellite
    # 5 estimated reflector height
    # 6 normalized LSP amplitude
    # 7 normalized LS amplitude, 
    # 8 hour of day, UTC
    # 9 raw LSP, for advanced setting
    # 10 raw LS amp, for advanced setting
    # 11 apriori RH
    # 12 quadrant (pboh2o style)
    # 13 delRH (for adv model)
    # 14 vegMask (for adv model)


    # this is the number of points for a given satellite track
    # just reassigning hte variable name
    reqNumpts = min_req_pts_track

    # KL disclosure
    # this code was written as a quick port between matlab and python
    # it does use many of the nice features of python variables. 
    
    # checking each geographic quadrant
    k4 = 1
    # two list of tracks - lets you compare an old run with a new run as a form of QC
    newlist = station + '_tmp.txt'
    oldlist = xdir + '/input/' + station + '_phaseRH.txt'

    ftmp = open(newlist,'w+')
    ftmp.write("{0:s} \n".format( '% station ' + station) )
    ftmp.write("{0:s} \n".format( '% TrackN  RefH SatNu MeanAz  Nval  Azimuths'))
    # open up a figure for the phase results by quadrant
    fig,ax = matplt.subplots(2, 2, figsize=(10,10))
    matplt.suptitle(f"Station: {station}", size=12)

    # open up a second plot for the amplitudes in the advanced option
    if advanced: 
        # opens the file , writes a header
        #fname_phase = f'{xdir}/Files/{subdir}/{station}_{str(year)}_all_phase.txt'

        fig2,ax2 = matplt.subplots(2, 2, figsize=(10,10))
        matplt.suptitle(f"Lomb Scargle Periodogram Amplitudes: {station}", size=12)

    for index, az in enumerate(azlist):
        b = 0
        k += 1
        ww = 0
        print('quadrant ' , oldquads[index])
        amin = az ; amax = az + 90
        # make a quadrant average for plotting purposes
        vquad = np.empty(shape=[0, 4])
        # pick up the sat list from the actual list
        satlist = stracks[atracks == amin]

        # set the titles for the two plots
        ax[bx[index],by[index]].set_title(f'Azimuth {str(amin)}-{str(amax)} deg.',fontsize=fs)
        if advanced:
            ax2[bx[index],by[index]].set_title(f'Azimuth {str(amin)}-{str(amax)} deg.',fontsize=fs)

        for satellite in satlist:
            # set the indices for the satellite and quadrant you want to look at here
            ii = (ssat == satellite) & (azdata > amin) & (azdata < amax) & (phase < 360)
            y,t,h,x,azd,s,amp_lsps,amp_lss,rhs,ap_rhs = \
                    qp.rename_vals(year_sat_phase, doy, hr, phase, azdata, ssat, amp_lsp, amp_ls, rh, ap_rh, ii)
            if screenstats:
                print('Looking at ', int(satellite), amin, amax,' Num vals', len(y))

            iikk  = (atracks == amin) & (stracks == satellite) 
            rhtrack = float(tracks[iikk,1]) # a priori RH
            meanaztrack = float(tracks[iikk,3])
            nvalstrack = float(tracks[iikk,4])

            if len(x) > reqNumpts:
                b += 1 # is b used?
                sortY = np.sort(x)
                N = len(sortY)
                NN = int(np.round(0.20*N))
                # use median value instead
                medv = np.median(sortY[(N-NN):(N-1)])
                new_phase = -(x-medv)
                # this might be a problem ???? maybe use -30?
                ii = (new_phase > -20)

                y,t,h,new_phase,azd,s,amp_lsps,amp_lss,rhs,ap_rhs = \
                        qp.rename_vals(y, t, h, new_phase, azd, s, amp_lsps, amp_lss, rhs, ap_rhs,ii)

                if len(t) == 0:
                    print('you should consider removing this satellite track as there are no results', satellite, amin)

                if (len(t) > reqNumpts):
                    ww = ww + 1 # index for plotting in a quadrant

                    # not sure why this is done in a loop - and why it is even here????
                    # and why with len(t) - 1
                    for l in range(0, len(t)-1):
                        if new_phase[l] > 340:
                            new_phase[l] = new_phase[l] - 360

                    # this is ok for regular model - not so good for big vegetation sites
                    ii = (new_phase > -20) & (new_phase < 60)
                    # just wondering
                    if advanced:
                        ii = (new_phase > -30) & (new_phase < 100)

                    y,t,h,new_phase,azd,s,amp_lsps,amp_lss,rhs,ap_rhs = \
                            qp.rename_vals(y, t, h, new_phase, azd, s, amp_lsps, amp_lss, rhs, ap_rhs, ii)

                    sortY = np.sort(new_phase)
                    # looks like I am using the bottom 20% 
                    NN = int(np.round(0.2*len(sortY)))
                    mv = np.median(sortY[0:NN])
                    new_phase = new_phase - mv
                    fracyear = y + t/365.25

                    newl = np.vstack((y, t, new_phase, azd)).T
                    vquad = np.vstack((vquad, newl))
                    if screenstats:
                        print('here0')

                    # this is to normalize the amplitudes. use base 15% to set it
                    basepercent = 0.15
                    # these are normalized LSP amplitudes
                    norm_ampLSP = qp.normAmp(amp_lsps, basepercent) ; 
                    # these are normalized LS amplitudes
                    norm_ampLS= qp.normAmp(amp_lss, basepercent)

                    # adding three new columns to use in Clara Chew algorithm
                    NN = len(amp_lsps)
                    qs = oldquads[index]*np.ones(shape=[1,NN])
                    delRH = rhs-ap_rhs
                    i = (norm_ampLSP < 0.8)
                    vegMask = np.zeros(shape=[NN,1])
                    vegMask[i] = 1

                    newl2 = np.vstack((y, t, new_phase, azd, s, rhs, norm_ampLSP,norm_ampLS,h,amp_lsps,amp_lss,ap_rhs,qs,delRH,vegMask.T)).T
                    #newl2 = np.vstack((y, t, new_phase, azd, s, rhs, norm_ampLSP,norm_ampLS,h,amp_lsps,amp_lss,ap_rhs,qs)).T

                    newl = np.vstack((y, t, new_phase, azd, s, rhs, norm_ampLSP,norm_ampLS,h,amp_lsps,amp_lss,ap_rhs)).T

                    # this is a kind of quality control -use previous solution to have 
                    # better feel for whether current solution works. defintely needs to go in a function

                    if (len(newl) > reqNumpts): 
                        k4= qp.kinda_qc(satellite, rhtrack,meanaztrack,nvalstrack, amin,amax, y, t, new_phase, 
                                         avg_date,avg_phase,warning_value,ftmp,remove_bad_tracks,k4,avg_exist )
                    else:
                        print('No previous solution or not enough points for this satellite.', int(satellite), amin, amax,len(newl))

                    adv_color = colors[ww:ww+1] # sets color for below
                    # stack this latest set of values to vxyz
                    #vxyz = np.vstack((vxyz, newl))
                    # try using version with quadrants
                    vxyz = np.vstack((vxyz, newl2))
                    datetime_dates = []
                    csat = str(int(satellite))
                    # make datetime dates for x-axis plots
                    for yr, d in zip(y, t):
                        datetime_dates.append(datetime.strptime(f'{int(yr)} {int(d)}', '%Y %j'))

                    if advanced:
                        ax[bx[index],by[index]].plot(datetime_dates, new_phase, 'o', markersize=3,color=adv_color,label=csat)
                    else:
                        ax[bx[index],by[index]].plot(datetime_dates, new_phase, 'o', markersize=3,label=csat)

                    # per clara chew paper in GPS Solutions 2016
                    if advanced:
                        # sort for the smoothing ... cause ... you can imagine 
                        ik = np.argsort( fracyear)
                        try:
                            smoothAmps = scipy.signal.savgol_filter(norm_ampLSP[ik], window_length=31,polyorder=2 )
                            # turn off for now - changed x-axis to datetime instead of fractional year
                            #ax2[bx[index],by[index]].plot(fracyear[ik], smoothAmps, '-',color=adv_color)
                        except:
                            print('some issue with the smoothing')
                        # not sure this will work
                        ax2[bx[index],by[index]].plot(datetime_dates, norm_ampLSP, '.',color=adv_color,label=csat)
                        #ax2[bx[index],by[index]].plot(fracyear[ik], norm_ampLSP[ik], '.',color=adv_color,label=csat)


        # now add things to the plots for the whole quadrant, like labels and grid lines
        if (index == 0 ) or (index == 2):
            ax[bx[index],by[index]].set_ylabel('Phase')
            if advanced:
                ax2[bx[index],by[index]].set_ylabel('NormAmps')

        # for phase
        ax[bx[index],by[index]].set_ylim((-20,60))
        if advanced:
            ax[bx[index],by[index]].set_ylim((-30,90))

        ax[bx[index],by[index]].grid()
        if sat_legend and (ww > 0):
            ax[bx[index],by[index]].legend(loc='upper right',fontsize=fs-2)

        fig.autofmt_xdate() # set for datetime

        # for normalized amplitude plot, only triggered by advanced setting
        if advanced:
            fig2.autofmt_xdate() # set for datetime
            ax2[bx[index],by[index]].grid()
            if sat_legend and (ww > 0):
                ax2[bx[index],by[index]].legend(loc='upper right',fontsize=fs-2)

    ftmp.close()

    if remove_bad_tracks:
        print('Writing out a new list of good satellite tracks to ', oldlist)
        subprocess.call(['mv','-f', newlist, oldlist])
    else:
        subprocess.call(['rm','-f', newlist ])

    qp.save_vwc_plot(fig,  f'{xdir}/Files/{subdir}/{station}_az_phase.png')

    if advanced:
        qp.save_vwc_plot(fig2,  f'{xdir}/Files/{subdir}/{station}_az_normamp.png')


    #Need to Define clearly the stored values in vxyz

    if advanced:
        print('Still working on the advanced option. Exiting.')
        fname_phase = f'{xdir}/Files/{subdir}/{station}_all_phase.txt'
        qp.write_phase_for_advanced(fname_phase, vxyz)

        if plt:
            matplt.show()
        sys.exit()
    else:
        # write out daily phase values
        tv = qp.write_avg_phase(station, phase, fr,year,year_end,minvalperday,vxyz,subdir)
        print('Number of daily phase measurements ', len(tv))
        if len(tv) < 1:
            print('No results - perhaps minvalperday or min_req_pts_track are too stringent')
            sys.exit()

        # make datetime date array
        datetime_dates = [datetime.strptime(f'{int(yr)} {int(d)}', '%Y %j') for yr, d in zip(tv[:, 0], tv[:, 1])]

        # make a plot of daily phase values
        qp.daily_phase_plot(station, fr,datetime_dates, tv,xdir,subdir,hires_figs)

        # convert daily phase values to volumetric water content
        qp.convert_phase(station, year, year_end, plt,fr,tmin,tmax,polyorder,circles,subdir,hires_figs)


def main():
    args = parse_arguments()
    vwc(**args)


if __name__ == "__main__":
    main()



import os 
import subprocess
import sys
import time
import wget

import gnssrefl.gps as g

def cddis_highrate(station, year, month, day,stream,dec_rate):
    """
    picks up highrate RINEX files from CDDIS
    and merges them

    Parameters
    ----------
    station : str
        4 char or 9 char station name
        Rinex 2.11 for the first and rinex 3 for the latter
    year : int
        full year
    month : int
        calendar month
    day : int 
        calendar day
    stream : str
        rinex3 ID, S or R
    dec_rate : int
        decimation rate, seconds

    Returns
    -------
    rinexname : str
        name of the merged/uncompressed outputfile
    fexist : bool
        whether the Rinex file was successfully retrieved

    requires hatanaka code and gfzrnx
    """
    fexist  = False
    if len(station) == 4:
        version = 2
    else:
        version = 3
    crnxpath = g.hatanaka_version()
    gfzpath = g.gfz_version()
    alpha='abcdefghijklmnopqrstuvwxyz'
    # if doy is input
    if day == 0:
        doy=month
        d = g.doy2ymd(year,doy);
        month = d.month; day = d.day
    doy,cdoy,cyyyy,cyy = g.ymd2doy(year,month,day); 

    if not os.path.isfile(gfzpath):
        print('You need to install gfzrnx to use high-rate RINEX data in my code.')
        return

    gns = 'https://cddis.nasa.gov/archive/gnss/data/highrate/' 
    gns = gns + cyyyy + '/'+ cdoy + '/' +cyy + 'd/'
    #YYYY/DDD/YYt/HH/mmmmDDDHMM.YYt.gz

    s1=time.time()
    print('WARNING: Get yourself a cup of coffeee. Downloading 96 files takes a long time.')
    fileF = 0
    streamID  = '_' + stream + '_'
    s1 = time.time()
    for h in range(0,24):
        # subdirectory
        ch = '{:02d}'.format(h)
        print('Hour: ', ch)
        for e in ['00', '15', '30', '45']:
            if version == 2:
                oname = station + cdoy + alpha[h] + e + '.' + cyy + 'o'; 
                file_name, crnx_name, file_name2, crnx_name2, exe1, exe2 = variableArchives(station,year,doy,cyyyy,cyy,cdoy,alpha[h],e) 
            else:
                file_name = station.upper() + streamID + cyyyy + cdoy + ch + e + '_15M_01S_MO.crx.gz'
                crnx_name = file_name[:-3] 
                oname = station.upper() + streamID + cyyyy + cdoy + ch + e + '_15M_01S_MO.rnx' # do we need this?

            new_way_dir = '/gnss/data/highrate/' + cyyyy + '/' + cdoy + '/' + cyy + 'd/' + ch + '/'
            #print(new_way_dir)
            if os.path.isfile(oname):
                print('Found it:', new_way_dir,file_name)
                fileF = fileF + 1
            elif os.path.isfile(file_name):
                print('Found gzip/hatanaka file:', new_way_dir,file_name)
                subprocess.call(['gunzip',file_name])
                subprocess.call([crnxpath, crnx_name])
                subprocess.call(['rm',crnx_name])
                fileF = fileF + 1
            else:
                print('Looking for:', new_way_dir,file_name)
                try:
                    g.cddis_download_2022B(file_name,new_way_dir)
                    if (version == 3):
                        if os.path.isfile(file_name): 
                            subprocess.call(['gunzip',file_name])
                            subprocess.call([crnxpath, crnx_name])
                            subprocess.call(['rm',crnx_name])
                    if (version == 2):
                        if os.path.isfile(file_name):
                            subprocess.call([exe1,file_name])
                            subprocess.call([crnxpath, crnx_name])
                            subprocess.call(['rm',crnx_name])
                        else:
                            g.cddis_download_2022B(file_name2,new_way_dir)
                            subprocess.call([exe2,file_name2])
                            subprocess.call([crnxpath, crnx_name2])
                            subprocess.call(['rm',crnx_name2])
                except:
                    #print('Failure using cddis_download_2022B')
                    subprocess.call(['rm','-f',file_name])

                if os.path.isfile(oname):
                    fileF = fileF + 1
                    print(' >>>  and found it')
    if version == 2:
        searchpath = station + cdoy + '*.' + cyy + 'o'
        rinexname = station + cdoy + '0.' + cyy + 'o'
        tmpname = station + cdoy + '0.' + cyy + 'o.tmp'
    else:
        searchpath = station.upper() + streamID + cyyyy + cdoy + '*.rnx'
        rinexname = station.upper() + streamID + cyyyy + cdoy + '0000_01D_01S_MO.rnx'
        tmpname = rinexname + 'tmp'

    s2=time.time()
    print('That download experience took ', int(s2-s1), ' seconds.')
    print('Attempt to merge the 15 minute files using gfzrnx and move to ', rinexname)
    if (fileF > 0): # files exist
        if (dec_rate == 1):
            subprocess.call([gfzpath,'-finp', searchpath, '-fout', tmpname, '-vo',str(version),'-f','-q'])
        else:
            s3=time.time()
            crate = str(dec_rate)
            subprocess.call([gfzpath,'-finp', searchpath, '-fout', tmpname, '-vo',str(version),'-sei','out','-smp',crate,'-f','-q'])
            s4=time.time()

        cm = 'rm ' + searchpath 
        if os.path.isfile(tmpname): # clean up
            ok = 1
            subprocess.call(cm,shell=True)
            subprocess.call(['mv',tmpname,rinexname])
            print('File created ', rinexname)
            fexist = True

    s2=time.time()
    print('The whole experience took ', int(s2-s1), ' seconds.')
    return rinexname,  fexist

def variableArchives(station,year,doy,cyyyy,cyy, cdoy,chh,cmm):
    """
    creates rinex3 compliant file names and finds executables needed
    to manipulate those files

    Parameters
    ----------
    station : str
        9 ch station name
    year : int
        full year
    doy : int
        day of year
    cyyyy : str
        4 ch year
    cyy : str
        two ch year
    cdoy : str
        3 ch day of year
    chh : str
        2 ch hour
    cmm : str
        2 ch minutes

    Returns
    -------
    file_name : str 
        first filename to look for 
    crnx_name : str
        first hatanaka name 
    file_name2 : str
        second filename to look for
    crnx_name2 : str
        second hatanaka compressed name 
    exe1 : str
        uncompression executable to use for file_name
    exe2 : str
        uncompression executable to use for file_name2

    """
# Before being merged into tar files, all Unix compressed RINEX V2 data with file 
#  extension ".Z" will be switched to gzip compression with the file extension ".gz". This 
#  change in compression is in accordance with the IGS transition to gzip conversion for 
#  RINEX V2 data after December 1, 2020.
#
# 335 is december 1 2020
    T0 = 2020 + 335/365.25
    if ((year+doy/365.25) <= T0):
        file_name = station + cdoy + chh + cmm + '.' + cyy + 'd.Z'
        crnx_name = file_name[:-2]
        file_name2 = station + cdoy + chh + cmm + '.' + cyy + 'd.gz'
        crnx_name2 = file_name2[:-3]
        exe1 = 'uncompress'
        exe2 = 'gunzip'
    else:
        file_name = station + cdoy +  chh + cmm + '.' + cyy + 'd.gz'
        crnx_name = file_name[:-3]
        file_name2 = station + cdoy + chh + cmm + '.' + cyy + 'd.Z'
        crnx_name2 = file_name2[:-2]
        exe2 = 'uncompress'
        exe1 = 'gunzip'

    return file_name, crnx_name, file_name2, crnx_name2, exe1, exe2

def bkg_highrate(station, year, month, day,stream,dec_rate,bkg,**kwargs):
    """
    picks up a highrate RINEX 3 file from BKG, merges and decimates it.
    requires gfzrnx

    Parameters
    ----------
    station : str
        9 ch station name 
    year : int
        full year
    month : integer
        month or day of year if day set to 0
    day : int
        day of the month
    stream : str
        R or S
    dec_rate : int
        decimation rate in seconds
    bkg : str
        file directory at BKG (igs or euref)

    Returns
    -------
    file_name24 : str
        name of merged rinex file
    fexist : boolean
        whether file exists

    """
    timeout = kwargs.get('timeout',0)
    if (timeout > 0) :
        print('timeout parameter has been set')

    fexist  = False
    version = 3
    crnxpath = g.hatanaka_version()
    gexe = g.gfz_version()
    alpha='abcdefghijklmnopqrstuvwxyz'
    # if doy is input
    if day == 0:
        doy=month
        d = g.doy2ymd(year,doy);
        month = d.month; day = d.day
    doy,cdoy,cyyyy,cyy = g.ymd2doy(year,month,day); 

    if not os.path.isfile(crnxpath):
        g.hatanaka_warning(); return

    if not os.path.isfile(gexe):
        print('You need to install gfzrnx to use high-rate RINEX data in my code.')
        return '', fexist

#    https://igs.bkg.bund.de/root_ftp/EUREF/highrate/2022/233/a/VLIS00NLD_R_20222330000_15M_01S_MO.crx.gz
    gns = 'https://igs.bkg.bund.de/root_ftp/' + bkg + '/highrate/'
    # base directory name
    gns = gns + cyyyy + '/'+ cdoy + '/' 
    print('looking for files in: ', gns)

    s1=time.time()
    print('WARNING: Get yourself a cup of coffeee. Downloading 96 files takes a long time.')
    fileF = 0
    streamID  = '_' + stream + '_'
    s1 = time.time()
    for h in range(0,24):
        # subdirectory
        ch = '{:02d}'.format(h)
        print('Hour: ', ch)
        for e in ['00', '15', '30', '45']:
            file_name = station.upper() + streamID + cyyyy + cdoy + ch + e + '_15M_01S_MO.crx.gz'
            dirname = gns + '/' + alpha[h] + '/'
            #print('looking for', dirname + file_name)
            crnx_name = file_name[:-3] 
            oname = file_name[:-6] + 'rnx'

            if os.path.isfile(oname):
                fileF = fileF + 1
                print('You already have ', oname, ' so no need to download')
            else:
                try:
                    if timeout > 0:
                        #print('using replace_wget')
                        s = g.replace_wget(dirname+file_name, file_name,timeout=timeout)
                    else:
                        # it is not getting all the files. maybe go back to wget
                        #s = g.replace_wget(dirname+file_name, file_name)
                        #print('using wget.download')
                        wget.download(dirname+file_name,file_name)
                    if os.path.isfile(file_name):
                        subprocess.call(['gunzip',file_name]) # unzip
                        subprocess.call([crnxpath, crnx_name]) # hatanaka
                        subprocess.call(['rm',crnx_name]) # remove old file
                except:
                    okok = 1
                    print('Could not find ', file_name, ' at BKG')
                if os.path.isfile(oname):
                    #print('successful download ', oname)
                    fileF = fileF + 1
                else:
                    print('Unsuccessful download ', oname)

    searchP = station.upper() + streamID + cyyyy + cdoy + '*15M*MO.rnx'
    print('Found ', fileF,' 15 minute files')

    outfile = station.upper() + cyyyy + cdoy + '.tmp'
    crate = '{:02d}'.format(dec_rate)

    file_name24 = ''

    if (fileF > 0):
        subprocess.call([gexe,'-finp', searchP, '-fout', outfile, '-vo','3', '-smp', crate, '-f','-q'])
        file_name24 = station.upper() + streamID + cyyyy + cdoy + '0000_01D_' + crate + 'S_MO.rnx'
        subprocess.call(['mv',outfile, file_name24]) # remove old file
        fexist = True

    s2=time.time()
    print('That download and merging experience took ', int(s2-s1), ' seconds.')

    # remove 15 minute files
    cm = 'rm ' + station.upper() + streamID + cyyyy + cdoy + '*15M_01S_MO.rnx'
    if fexist:
        subprocess.call(cm,shell=True)

    return file_name24,  fexist


def esp_highrate(station, year, month, day,stream,dec_rate):
    """
    picks up a highrate RINEX 3 file from Spanish Geodeic Center, merges and decimates it.
    requires gfzrnx

    Parameters
    ----------
    inputs: string
        9 ch station name 
    year : integer
        full year
    month : integer
        month or day of year if day set to 0
    day : integer
        day of the month
    stream : str
        R or S
    dec_rate : integer
        decimation rate in seconds

    Returns
    -------
    file_name24 : str
        name of merged rinex file
    fexist : boolean
        whether file exists

    """
    fexist  = False
    version = 3
    crnxpath = g.hatanaka_version()
    gexe = g.gfz_version()
    # if doy is input
    if day == 0:
        doy=month
        d = g.doy2ymd(year,doy);
        month = d.month; day = d.day
    doy,cdoy,cyyyy,cyy = g.ymd2doy(year,month,day); 

    if not os.path.isfile(crnxpath):
        g.hatanaka_warning(); return

    if not os.path.isfile(gexe):
        print('You need to install gfzrnx to use high-rate RINEX data in my code.')
        return '', fexist

    fileid = str(year) + '{:02d}'.format(month) + '{:02d}'.format(day)

#    https://igs.bkg.bund.de/root_ftp/EUREF/highrate/2022/233/a/VLIS00NLD_R_20222330000_15M_01S_MO.crx.gz
#   https://datos-geodesia.ign.es/ERGNSS/horario_1s/20231110/00/
    gns = 'https://datos-geodesia.ign.es/ERGNSS/horario_1s/' + fileid + '/' 
    # base directory name
    print('looking for files in: ', gns)

    s1=time.time()
    print('WARNING: Downloading and merging 24 one hour files.')
    fileF = 0
    streamID  = '_' + stream + '_'
    s1 = time.time()
    for h in range(0,23):
        # subdirectory
        ch = '{:02d}'.format(h)
        print('Hour: ', ch)
        file_name = station.upper() + streamID + cyyyy + cdoy + ch + '00_01H_01S_MO.crx.gz'
        dirname = gns + '/' + ch + '/'
        #print('looking for', dirname + file_name)
        crnx_name = file_name[:-3] 
        oname = file_name[:-6] + 'rnx'
        if os.path.isfile(oname):
            fileF = fileF + 1
            print('already have ', oname)
        else:
            try:
                wget.download(dirname+file_name,file_name)
                subprocess.call(['gunzip',file_name]) # gunzip
                subprocess.call([crnxpath, crnx_name]) # hatanaka
                subprocess.call(['rm',crnx_name]) # remove old file
            except:
                okok = 1
            if os.path.isfile(oname):
                print('Successful download ', oname)
                fileF = fileF + 1
            else:
                print('Unsuccessful download ', oname)

    searchP = station.upper() + streamID + cyyyy + cdoy + '*01H*MO.rnx'
    print(searchP)
    print('Found ', fileF,' one hour files')

    outfile = station.upper() + '.tmp'
    crate = '{:02d}'.format(dec_rate)

    file_name24 = ''

    file_name24 = station.upper() + streamID + cyyyy + cdoy + '0000_01D_' + crate + 'S_MO.rnx'
    print(file_name24,outfile)
    if (fileF > 0):
        subprocess.call([gexe,'-finp', searchP, '-fout', outfile, '-vo','3', '-smp', crate, '-f','-q'])
        subprocess.call(['mv',outfile, file_name24]) # remove old file
        fexist = True

    s2=time.time()
    print('That download and merging experience took ', int(s2-s1), ' seconds.')

    # remove one hour files
    cm = 'rm ' + station.upper() + streamID + cyyyy + cdoy + '*01H_01S_MO.rnx'
    if fexist:
        subprocess.call(cm,shell=True)

    return file_name24,  fexist


def cddis_highrate_tar(station, year, month, day,stream,dec_rate):
    """
    picks up a tar'ed highrate RINEX files from CDDIS, untars it, and merges.

    Parameters
    ----------
    station : str
        4 char or 9 char station name
        Rinex 2.11 for the first and rinex 3 for the latter
    year : int
        full year
    month : int
        calendar month
    day : int 
        calendar day
    stream : str
        rinex3 ID, S or R
    dec_rate : int
        decimation rate, seconds

    Returns
    -------
    rinexname : str
        name of the merged/uncompressed outputfile
    fexist : bool
        whether the Rinex file was successfully retrieved

    requires hatanaka code and gfzrnx
    """
    s1=time.time()
    fexist  = False
    version = 3 # only version 3
    crnxpath = g.hatanaka_version()
    gfzpath = g.gfz_version()
    alpha='abcdefghijklmnopqrstuvwxyz'
    # if doy is input
    if day == 0:
        doy=month
        d = g.doy2ymd(year,doy);
        month = d.month; day = d.day
    doy,cdoy,cyyyy,cyy = g.ymd2doy(year,month,day); 

    if not os.path.isfile(gfzpath):
        print('You need to install gfzrnx to use high-rate RINEX data in gnssrefl.')
        return

    gns = 'https://cddis.nasa.gov/archive/gnss/data/highrate/' 
    gns = gns + cyyyy + '/'+ cdoy + '/' +cyy + 'd/'

    streamID  = '_' + stream + '_'

    new_way_dir = 'gnss/data/highrate/' + cyyyy + '/' + cdoy + '/'
    file_name = station.upper() +  streamID + cyyyy + cdoy + '0000_01D_01S_MO.crx.tar'
    #print(new_way_dir,file_name)
    if os.path.isfile(file_name):
        print('already have the tar file - so it has probably been un tarred')
    else:
        g.cddis_download_2022B(file_name,new_way_dir)
        if os.path.isfile(file_name):
            subprocess.call(['tar','-xf', file_name])
            print('Now remove the tar file')
            subprocess.call(['rm', file_name])

    fileF = 0
    for h in range(0,24):
        # subdirectory
        ch = '{:02d}'.format(h)
        print('Hour: ', ch)
        #gnss/data/highrate/2023/199/23d/22
        new_way_dir = 'gnss/data/highrate/' + cyyyy + '/' + cdoy + '/' + cyy + 'd/' + ch + '/'
        for e in ['00', '15', '30', '45']:
            file_name = station.upper() + streamID + cyyyy + cdoy + ch + e + '_15M_01S_MO.crx.gz'
            crnx_name = file_name[:-3] 
            oname = station.upper() + streamID + cyyyy + cdoy + ch + e + '_15M_01S_MO.rnx' # do we need this?
            if os.path.isfile(new_way_dir + oname):
                #print('Found rnx file', oname)
                subprocess.call(['mv',new_way_dir + oname, 'gnss/data'])
                fileF = fileF + 1
            elif os.path.isfile(new_way_dir + file_name):
                #print('Found file:', new_way_dir,file_name)
                subprocess.call(['gunzip',new_way_dir + file_name])
                subprocess.call([crnxpath, new_way_dir + crnx_name])
                subprocess.call(['rm',new_way_dir + crnx_name])
                subprocess.call(['mv',new_way_dir + oname, 'gnss/data'])
                if os.path.isfile('gnss/data/' + oname):
                    fileF = fileF + 1
            else:
                print('did not find  file:', new_way_dir,file_name)
    # moved the files to 'gnss/data/'
    searchpath = 'gnss/data/' + station.upper() + streamID + cyyyy + cdoy + '*.rnx'
    #print(searchpath)
    rinexname = station.upper() + streamID + cyyyy + cdoy + '0000_01D_01S_MO.rnx'
    tmpname = rinexname + 'tmp'

    print('Attempt to merge the 15 minute files using gfzrnx and move to ', rinexname)
    if (fileF > 0): # files exist
        if (dec_rate == 1):
            subprocess.call([gfzpath,'-finp', searchpath, '-fout', tmpname, '-vo',str(version),'-f','-q'])
        else:
            print('Am decimating with gfzrnx')
            crate = str(dec_rate)
            subprocess.call([gfzpath,'-finp', searchpath, '-fout', tmpname, '-vo',str(version),'-sei','out','-smp',crate,'-f','-q'])

        cm = 'rm ' + searchpath 
        if os.path.isfile(tmpname): # clean up
            ok = 1
            subprocess.call(cm,shell=True)
            subprocess.call(['mv',tmpname,rinexname])
            print('File created ', rinexname)
            fexist = True

        # remove unneeded 15 minute files
        #new_way_dir = 'gnss/data/highrate/' + cyyyy + '/' + cdoy + '/' + cyy + 'd/' + ch + '/'
        new_dir =      'gnss/data/highrate/' + cyyyy + '/' + cdoy + '/' + cyy + 'd/*'
        print('remove compressed 15 minute files')
        cm = 'rm -rf ' + new_dir
        subprocess.call(cm,shell=True)

        print('remove converted 15 minute files')
        searchpath = 'gnss/data/' + station.upper() + streamID + cyyyy + cdoy + '*.rnx'
        cm = 'rm -rf ' + searchpath
        subprocess.call(cm,shell=True)

        # remove junk from CDDIS
        searchpath = station.upper() + streamID + cyyyy + cdoy + '*.md5.txt'
        cm = 'rm -rf ' + searchpath
        subprocess.call(cm,shell=True)

        searchpath = station.upper() + streamID + cyyyy + cdoy + '*.sha512.txt'
        cm = 'rm -rf ' + searchpath
        subprocess.call(cm,shell=True)

    s2=time.time()
    print('That download/merge experience took ', int(s2-s1), ' seconds.')
    return rinexname,  fexist

def bkg_highrate_tar(station, year, month, day,stream,dec_rate,bkg):
    """
    picks up a highrate RINEX 3 file from BKG, merges and decimates it.
    requires gfzrnx

    Parameters
    ----------
    station : str
        9 ch station name 
    year : int
        full year
    month : integer
        month or day of year if day set to 0
    day : int
        day of the month
    stream : str
        R or S
    dec_rate : int
        decimation rate in seconds
    bkg : str
        file directory at BKG (igs or euref)

    Returns
    -------
    file_name24 : str
        name of merged rinex file
    fexist : boolean
        whether file exists

    """
    fexist  = False
    version = 3
    crnxpath = g.hatanaka_version()
    gexe = g.gfz_version()
    alpha='abcdefghijklmnopqrstuvwxyz'
    # if doy is input
    if day == 0:
        doy=month
        d = g.doy2ymd(year,doy);
        month = d.month; day = d.day
    doy,cdoy,cyyyy,cyy = g.ymd2doy(year,month,day); 

    if not os.path.isfile(crnxpath):
        g.hatanaka_warning(); return

    if not os.path.isfile(gexe):
        print('You need to install gfzrnx to use high-rate RINEX data in my code.')
        return '', fexist

#    https://igs.bkg.bund.de/root_ftp/EUREF/highrate/2022/233/a/VLIS00NLD_R_20222330000_15M_01S_MO.crx.gz
# 	VLIS00NLD_R_20231220000_01D_01S_MO.crx.tar.gz
    gns = 'https://igs.bkg.bund.de/root_ftp/' + bkg + '/highrate/'
    streamID  = '_' + stream + '_'
    # base directory name
    dirname = gns + cyyyy + '/'+ cdoy + '/' 
    xend = '0000_01D_01S_MO'
    #print('looking for files in: ', dirname)
    file_name = station.upper() + streamID + cyyyy + cdoy + xend + '.crx'
    file_name1 = file_name + '.tar.gz'
    print(dirname + file_name1)
    if not os.path.isfile(file_name + '.tar'):
        s1=time.time()
        s = g.replace_wget(dirname+file_name1, file_name1)
        s2=time.time()
        print('That file took ', int(s2-s1), ' seconds to download')
        if os.path.isfile(file_name1):
            subprocess.call(['gunzip',file_name1]) # unzip
            subprocess.call(['tar','-xf', file_name + '.tar'])
        else:
            print('Something is wrong with tar download')
            return '', False

    fileF = 0
    for h in range(0,24):
        # subdirectory
        ch = '{:02d}'.format(h)
        print('Hour: ', ch)
        for e in ['00', '15', '30', '45']:
            crnx_name = station.upper() + streamID + cyyyy + cdoy + ch + e + '_15M_01S_MO.crx'
            oname = crnx_name[:-3] + 'rnx'
            print(oname)

            if os.path.isfile(oname):
                fileF = fileF + 1
                print('You already have ', oname, ' so no need to make it ')
                if os.path.isfile(crnx_name):
                    subprocess.call(['rm',crnx_name])  
            else:
                if os.path.isfile(crnx_name):
                    subprocess.call([crnxpath,crnx_name])  
                    if os.path.isfile(oname):
                        fileF = fileF + 1
                        subprocess.call(['rm',crnx_name])  
                else:
                    print('File does not exist', oname)

    searchP = station.upper() + streamID + cyyyy + cdoy + '*15M*MO.rnx'
    print('Found ', fileF,' 15 minute files')


    outfile = station.upper() + streamID + cyyyy + cdoy + '.tmp'
    crate = '{:02d}'.format(dec_rate)

    file_name24 = station.upper() + streamID + cyyyy + cdoy + '0000_01D_01S_MO.rnx'
    print('Final name of RINEX 3 file ', file_name24)

    if os.path.isfile(outfile):
        print('you already merged')
        fexist = True
    else:
        if (fileF > 0):
            s1= time.time()
            subprocess.call([gexe,'-finp', searchP, '-fout', outfile, '-vo','3', '-smp', crate, '-f','-q'])
            fexist = True
            s2=time.time()
            print('That merging/decimating experience took ', int(s2-s1), ' seconds.')
            print('remove 15 minute files')
            cm = 'rm ' + station.upper() + streamID + cyyyy + cdoy + '*15M_01S_MO.rnx'
            subprocess.call(cm,shell=True)

    if fexist:
        print('Now change the output name: ', file_name24)
        subprocess.call(['mv', outfile, file_name24])
        subprocess.call(['rm', file_name+'.tar'])
    else:
        file_name24 = ''

    return file_name24,  fexist



import argparse
import numpy as np
import matplotlib.pyplot as mplt

import os
import subprocess
import sys

import gnssrefl.gps as g
import gnssrefl.subdaily as t
import gnssrefl.sd_libs as sd

from gnssrefl.utils import str2bool


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("station", help="station name", type=str)
    parser.add_argument("year", default=None, type=int, help="year of data")
    parser.add_argument("-year_end", default=None, type=int, help="Allow multiple years of inputs by specifying last year")
    parser.add_argument("-txtfile_part1", default=None, type=str, help="optional filename (part1), must be in gnssir output format") 
    parser.add_argument("-txtfile_part2", default=None, type=str, help="optional filename (part2), must be gnssir output format ") 
    parser.add_argument("-csv", default=None, type=str, help="set to True if you prefer csv to plain txt output")
    parser.add_argument("-plt", default=None, type=str, help="set to False to suppress plots")
    parser.add_argument("-spline_outlier1", default=None, type=float, help="outlier criterion used in first splinefit (meters)")
    parser.add_argument("-spline_outlier2", default=None, type=float, help="outlier criterion used in second splinefit (meters)")
    parser.add_argument("-knots", default=None, type=int, help="Knots per day, spline fit only (default is 8)")
    parser.add_argument("-sigma", default=None, type=float, help="part1 outlier criterion, float, default is 2.5 (sigma)")
    parser.add_argument("-extension", default=None, type=str, help="solution subdirectory")
    parser.add_argument("-rhdot", default=None, type=str, help="set to False if you want to stop after section 1 of the QC code")
    parser.add_argument("-doy1", default=None, type=int, help="initial day of year")
    parser.add_argument("-doy2", default=None, type=int, help="end day of year")
    parser.add_argument("-testing", default=None, type=str, help="set to False for old code ")
    parser.add_argument("-ampl", default=None, type=float, help="new amplitude constraint")
    parser.add_argument("-azim1", default=None, type=int, help="new min azimuth")
    parser.add_argument("-azim2", default=None, type=int, help="new max azimuth")
    parser.add_argument("-h1", default=None, type=float, help="min RH (m)")
    parser.add_argument("-h2", default=None, type=float, help="max RH (m)")
    parser.add_argument("-peak2noise", default=None, type=float, help="new peak2noise constraint")
    parser.add_argument("-kplt", default=None, type=str, help="special plot for kristine")
    parser.add_argument("-subdir", default=None, type=str, help="non-default subdirectory for output")
    parser.add_argument("-delta_out", default=None, type=int, help="Output interval for final spline fit, seconds (default is 1800)")
    parser.add_argument("-if_corr", default=None, type=str, help="Whether interfrequency correction is applied, optional")
    parser.add_argument("-knots_test", default=None, type=int, help="test knots, kristine only")
    parser.add_argument("-hires_figs", default=None, type=str, help="hi-resolution eps figures, default is False")
    parser.add_argument("-apply_rhdot", default=None, type=str, help="apply rhdot, default is True")
    parser.add_argument("-fs", default=None, type=int, help="fontsize for figures. default is 10")
    parser.add_argument("-alt_sigma", default=None, type=str, help="boolean test for alternate Nievinski sigma definition. default is False")
    parser.add_argument("-gap_min_val", default=None, type=float, help="min gap (hours) allowed in splinefit output file. default is 6 hours")
    parser.add_argument("-knots2", default=None, type=int, help="Secondary knots value for final fit. default is to use original knots value.")
    parser.add_argument("-gap_flag", default=None, type=str, help="if set to True/T, write 999 to spline fit for gaps")

    g.print_version_to_screen()

    args = parser.parse_args().__dict__

    # convert all expected boolean inputs from strings to booleans
    boolean_args = ['csv', 'plt', 'rhdot', 'testing','kplt','if_corr','hires_figs','apply_rhdot','alt_sigma','gap_flag']
    args = str2bool(args, boolean_args)

    # only return a dictionary of arguments that were added from the user - all other defaults will be set in code below
    return {key: value for key, value in args.items() if value is not None}


def subdaily(station: str, year: int, txtfile_part1: str = '', txtfile_part2: str = None, csv: bool = False, 
        plt: bool = True, spline_outlier1: float = None, spline_outlier2: float = None, 
        knots: int = None, sigma: float = None, extension: str = None, rhdot: bool = True, doy1: int = 1, 
        doy2: int = 366, testing: bool = True, ampl: float = None, h1: float=0.4, h2: float=300.0, 
        azim1: int=0, azim2: int = 360, peak2noise: float = 0, kplt: bool = False, 
        subdir: str = None, delta_out : int = None , if_corr: bool = True, knots_test: int = 0, 
             hires_figs : bool=False, apply_rhdot : bool=True, fs: int = 10, alt_sigma: bool= False, gap_min_val: float=6.0,
             year_end: int=None, knots2 : int=None, gap_flag: bool = False):
    """
    Subdaily combines gnssir solutions and applies relevant corrections needed to measure water levels (tides). 
    As of January 2024, it will allow multiple years. You can also specify which day of year to start with, i.e.
    -doy1 300 and -doy2 330 will do that range in a single year, or you could specific doy1 and doy2 as linked to 
    the start and stop year (year and year_end)

    In general this code is meant to be used at sites with tidal signals. If you have a site 
    without tidal signals, you should consider using daily_avg instead. If you would still 
    like to use this code for rivers and lakes, you should change the defaults for the spline 
    fits. For tidal sites, 8 knots per day is the default. For nearly stationary surfaces, as you 
    would expect for a lake or river, you should use many fewer knots per day.

    This code calculates and applies various corrections. New Reflector Height values are added 
    to the output files as new columns. If you run the code but continue to assume the "good answers" are
    in still in column 3, you are essentially not using the code at all. 

    As of version 3.1.4 you can use some of the subdaily optional inputs from the gnssir_input created json.
    See gnssir_input for details.


    As of version 2.0.0:

    The final output of subdaily is a smooth spline fit to reflector heights (RH) which has been 
    adjusted to mean sea level (meters). For this to be accurate, the user is asked to provide the orthometric height
    of the L1 GPS antenna phase center. This value should be stored as Hortho in the gnssir analysis strategy file
    (ssss.json where ssss is the 4 character station name). The output water levels are then defined as Hortho minus RH. 
    If the user does not provide Hortho, one is computed from the station ellipsoidal height stored in the gnssir 
    analysis strategy file and EGM96.

    The subdaily code has two main sections. 

    I. Summarize the retrievals (how many retrievals per constellation), identify and remove 
    gross outliers, provide plots to allow a user to evaluate Quality Control parameters. The 
    solutions can further be edited from the command line (i.e. restrict the RH using 
    -h1 and -h2, in meters, or azimuths using -azim1 and -azim2)

    II. This section has the following goals:

    - removes more outliers based on a spline fit to the RH retrievals

    - calculates and applies RHdot correction

    - removes an interfrequency (IF) bias. All solutions are then relative to GPS L1.

    txtfile_part1 is optional input if you want to skip concatenating daily gnssir output files 
    and use your own file. Make sure results are in the same format.

    txtfile_part2 is optional input that skips part 1 and uses this file as input to 
    the second part of the code.

    Examples
    --------
    subdaily at01 2023 -plt F
        for station at01, all solutions in 2023  but no plots to the screen

    subdaily at01 2023 -doy1 15 -doy2 45
        for all solutions in 2023 between days of year 15 through 45

    subdaily at01 2023 -h2 14 -if_corr F
        for all solutions in 2023 but with max RH set to 14 meters and interfrequency correction not applied 

    subdaily at01 2022 -year_end 2023
        analyze all data for years 2022 and 2023

    subdaily at03 2022 -azim1 180 -azim2 270
        restrict solutions to azimuths between 180 and 270

    Parameters
    ----------
    station : str
        4 character id of the station.
    year : int
        full year
    txtfile_part1 : str, optional
        input File name for part 1.
    txtfile_part2 : str, optional
        Input filename for part 2.
    csv: bool, optional
        Set to True if you would like csv in addition to plain txt.
        default is False.
    plt : bool, optional
        To print plots to screen or not.
        default is True.
    spline_outlier1 : float, optional
        Outlier criterion used in first splinefit, before RHdot  (m)
    spline_outlier2 : float, optional
        Outlier criterion used in second splinefit, after IF & RHdot (meters)
    knots : integer, optional
        Knots per day, spline fit only.
        default is 8.
    sigma : float, optional
        Simple sigma outlier criterion (e.g. 1 for 1sigma, 3 for 3sigma)
        default is 2.5
    extension : str, optional
        Solution subdirectory.
    rhdot : bool, optional
        Set to True to turn on spline fitting for RHdot correction.
        default is True.
    doy1 : int, optional
        Initial day of year, default is 1.
    doy2 : int, optional
        End day of year. Default is 366.
    testing : bool, optional
        Set to False for older code.
        default is now True.
    ampl : float, optional
        New amplitude constraint. Default is 0.
    azim1: int, optional
        minimum azimuth. Default is 0.
    azim2: int, optional
        Max azimuth. Default is 360.
    h1: float, optional 
        lowest allowed reflector height in meters. Default is 0.4
    h2: float, optional 
        highest allowed reflector height in meters. Default is 300
    peak2noise: float, optional
        New peak to noise constraint. Default is 0.
    kplt: bool, optional
        plot for kristine
    subdir : str, optional
        name for output subdirectory in REFL_CODE/Files
    delta_out : int, optional
        how frequently - in seconds - you want smooth spline model output written
        default is 1800 seconds
    if_corr : bool, option
        whether you want the inter-frequency removed
        default is true
    hires_figs : bool, optional
        whether high resolution figures are made
    apply_rhdot : bool, optional
        whether you want the RH dot correction applied
        for a lake or river you would not want it to be.
    fs : int, optional
        fontsize for Figures. default is 10 for now.
    alt_sigma : bool, optional
        whether you want to use Nievinski definition for outlier criterion.
        in part 1 of the code (the crude outlier detector)
    gap_min_val : float, optional
        removes splinefit values from output txt and plot for gaps 
        bigger than this value, in hours
    year_end : int, optional
        last year of analysis period.  
    knots2 : int, optional
        testing out allowing different knots for last spline
    gap_flag : bool
        whether you want gaps filled with 999 values in the final spline file 
    """

    print('<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>')
    print('<<<<<<<<<<<<<<<<<<<<<<<<<< subdaily >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>')
    print('<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>')

    if len(station) != 4:
        print('station names must be four characters long')
        sys.exit()

    # make surer environment variables are set
    g.check_environ_variables()
    xdir = os.environ['REFL_CODE']

    lsp = sd.pickup_subdaily_json_defaults(xdir, station, extension)

    if subdir is None: # not set on the command line
        if lsp['subdaily_subdir'] is None: 
            subdir = station
        else:
            subdir = lsp['subdaily_subdir']
            print('Using subdir value from json')

    # this makes sure the directory exists
    g.set_subdir(subdir)

    # I am not 100% sure of this - wish I had not created the subdir variable
    if extension is None:
        txtdir = xdir + '/Files/' + subdir
        extension = '' # is this needed? Yes, it is.
    else:
        txtdir = xdir + '/Files/' + subdir + '/' + extension

    g.checkFiles(station, extension)

    print('Using this directory for output: ', txtdir)

    
    # this is ugly - should be moved to a function

    if alt_sigma is None: # not set on the command line
        if lsp['subdaily_alt_sigma'] is not None: # but it has been set in the json
            alt_sigma = lsp['subdaily_alt_sigma']
            print('Using alt_sigma from json')


    if ampl is None: # not set on the command line
        if lsp['subdaily_ampl'] is not None: # but it has been set in the json
            ampl = lsp['subdaily_ampl']
            print('Using ampl value from json')
        else:
            ampl = 0 # don't go beyond whatever was done when you ran gnssir


    if delta_out is None: # not set on the command line
        if lsp['subdaily_delta_out'] is not None: # but it has been set in the json
            delta_out = lsp['subdaily_delta_out']
            print('Using delta_out from json')
        else:
            delta_out = 1800 # setting default here now instead of above

    if knots is None: # not set on the command line
        if lsp['subdaily_knots'] is not None: # but it has been set in the json
            knots = lsp['subdaily_knots']
            print('Using knots from json')
        else:
            knots = 8


    if sigma is None: # not set on the command line
        if lsp['subdaily_sigma'] is not None: # but it has been set in the json
            sigma = lsp['subdaily_sigma']
            print('Using sigma from json')
        else:
            sigma = 2.5

    if spline_outlier1 is None: # not set on the command line
        if lsp['subdaily_spline_outlier1'] is not None: # but it has been set in the json
            spline_outlier1 = lsp['subdaily_spline_outlier1']
            print('Using spline_outlier1 from json')

    if spline_outlier2 is None: # not set on the command line
        if lsp['subdaily_spline_outlier2'] is not None: # but it has been set in the json
            spline_outlier2 = lsp['subdaily_spline_outlier2']
            print('Using spline_outlier2 from json')

    #create the subdaily file
    if (h1 > h2):
        print('h1 must be less than h2. You submitted ', h1, ' and ', h2)
        sys.exit()

    outputs = [] # this is for multiple years

    if year_end is None: 
        year_end = year

    year_start = year

    multiyear = True
    if year_start == year_end:
        multiyear = False

    year_list = list(range(year, year_end+1))

    if txtfile_part2 is None:
        if txtfile_part1 == '':
            print('Will pick up and concatenate daily result files')
        else:
            print('Using ', txtfile_part1)
        # if txtfile provided, you can use that as your starting dataset 
        
        default_usage = True

        for y in range(year,year_end+1):
            dec31 = g.dec31(y)  # find doy for dec 31
            if not multiyear:
                doy_st = doy1
                doy_en = doy2
            else:
                if (y == year_end):
                    doy_st = 1
                    doy_en = doy2
                elif (y == year_start): 
                    doy_st = doy1
                    doy_en = dec31
                else:
                # in between year
                    doy_st = 1
                    doy_en = dec31

            print('Reading in the results for year: ', y, ' and doys ',  doy_st, ':', doy_en)

            ntv, obstimes, fname, fname_new = t.readin_and_plot(station, y, doy_st, doy_en, plt, \
                    extension, sigma, csv, azim1, azim2, ampl, peak2noise, txtfile_part1, \
                    h1,h2,kplt,txtdir,default_usage,hires_figs,fs,alt_sigma=alt_sigma)
            outputs.append(fname_new)

        haveObstimes = True
    else:
        haveObstimes = False
        fname_new = txtfile_part2
        if not os.path.isfile(fname_new):
            print('Input subdaily RH file you provided does not exist:', fname_new)
            sys.exit()

    # if you are not going to the second part of the code, display the plots, if requested. 
    if not rhdot:
        if plt:
            mplt.show()
    # only read in one year
    if year_end == year:
        input2spline = [fname_new]; output4spline = fname_new + '.withrhdot'

    # this is for multiyaer
    else:
        # this is a bit of a kluge - send it names of the yearly files
        input2spline = outputs 
        output4spline = txtdir + '/' + station + '_' + str(year) + '_' + str(year_end) + '.withrhdot'

    # not sure why tv and corr are being returned.
    if rhdot:
       tv, corr = t.rhdot_correction2(station, input2spline, output4spline, 
                                      plt, spline_outlier1, spline_outlier2, 
                                      knots=knots,txtdir=txtdir,testing=testing,delta_out=delta_out, 
                                      if_corr=if_corr,knots_test=knots_test,
                                      hires_figs=hires_figs, apply_rhdot=apply_rhdot,fs=fs,
                                      gap_min_val=gap_min_val,year=year, extension=extension, 
                                      knots2=knots2,csv=csv,gap_flag=gap_flag)
       if plt:
           mplt.show()

def main():
    args = parse_arguments()
    subdaily(**args)


if __name__ == "__main__":
    main()



import argparse
import numpy as np
import os
import subprocess
import sys

import gnssrefl.gps as g
import gnssrefl.refl_zones as rz
import gnssrefl.nyquist_libs as nl

from gnssrefl.utils import validate_input_datatypes, str2bool


def parse_arguments():
    parser = argparse.ArgumentParser()

    parser.add_argument('station', help='station', type=str)
    parser.add_argument('-lat', help='Latitude (deg), if station not in database', type=float, default=None)
    parser.add_argument('-lon', help='Longitude (deg), if station not in database', type=float, default=None)
    parser.add_argument('-el_height', help='Ellipsoidal height (m) if station not in database', type=float,default=None)
    parser.add_argument('-e1', help='min elevation angle (deg), default 5 ', type=float,default=None)
    parser.add_argument('-e2', help='max elevation angle (deg), default 25', type=float,default=None)
    parser.add_argument('-samplerate', help='receiver sample rate (sec), default 30', type=float,default=None)
    parser.add_argument('-system', help='default=gps, other options : galileo, glonass, beidou', type=str)
    parser.add_argument('-hires_figs', default=None,help='whether you want eps instead of png plots', type=str)

    args = parser.parse_args().__dict__

    # convert all expected boolean inputs from strings to booleans
    boolean_args = ['hires_figs']
    args = str2bool(args, boolean_args)


    return {key: value for key, value in args.items() if value is not None}

def max_resolve_RH(station: str, lat: float=None, lon: float=None, el_height: float=None, e1: float=5, e2: float=25, 
        samplerate : float = 30, system: str = 'gps', hires_figs : bool = False):
    """
    Calculates the Maximum Resolvable Reflector Height. This is analogous to a Nyquist
    frequency for GNSS-IR.  It creates a plot and makes a plain txt file in case you 
    want to look at the numbers.

    Examples
    --------
    max_resolve_RH sc02 -e1 5 -e2 15
        typical case for sites over water (5-15 degrees) but otherwise using defaults (GPS, 30 seconds)

    max_resolve_RH sc02 -samplerate 15 -system galileo
        Assume receiver sampling rate of 15 seconds and the Galileo constellation

    max_resolve_RH xxxx -lat 40 -lon 100 -el_height 20
        test your own site (xxxx) by inputting coordinates
    

    Parameters
    ----------
    station : str
        4 ch station name
    lat : float, optional
        latitude in deg
    lon : float, optional
        longitude in deg
    el_height : float, optional
        ellipsoidal height in m
    e1 : float
        min elevation angles (deg)
    e2 : float
        max elevation angles (deg)
    samplerate : float
        receiver sampling rate
    system : str, optional
        name of constellation (gps,glonass,galileo, beidou allowed)
        default is gps
    hires_figs : bool
        whether you want eps files instead of png 

    Returns
    -------
    Creates a figure file, stored in $REFL_CODE/Files/station

    """
    foundfiles = rz.save_reflzone_orbits()
    if not foundfiles:
        print('The orbit files needed for this code were either not found')
        print('or not downloaded successfully. They should be in the REFL_CODE/Files directory')
        sys.exit()



    if (lat is None) & (lon is None):
    # check the station coordinates in our database from the station name
        lat, lon, el_height = g.queryUNR_modern(station)
        if (lat == 0) and (lon == 0):
            print('Exiting.')
            sys.exit()
        else:
            print('Using inputs:', lat, lon, el_height)
    else:
        print('Using inputs:', lat, lon, el_height)

    orbfile, it =  rz.set_system(system)

    xx,yy,zz = g.llh2xyz(lat,lon,el_height);
    recv=np.array([xx,yy,zz])
    # main code that does everything
    nl.pickup_files_nyquist(station,recv,orbfile,it,e1,e2,samplerate,hires_figs)


def main():
    args = parse_arguments()
    data = max_resolve_RH(**args)


if __name__ == "__main__":
    main()




# subdaily libraries
import datetime
import math
import matplotlib.pyplot as plt
import numpy as np
from astropy.time import Time

# gnssrefl specific code
import gnssrefl.gps as g
import gnssrefl.gnssir_v2 as guts2

import scipy
import scipy.interpolate as interpolate
from scipy.interpolate import interp1d


def mjd_to_obstimes(mjd):
    """
    takes mjd array and converts to datetime for plotting.

    Parameters
    ----------
    mjd : numpy array of floats
        mod julian date

    Returns
    -------
    dt : numpy array of datetime objects

    """

    dt = Time(mjd,format='mjd').utc.datetime; 

    return dt


def write_spline_output(iyear, th, spline, delta_out, station, txtdir,Hortho):
    """
    Writing the output of the spline fit to the final RH time series.
    No output other than this text file for year 2023 and station name ssss:

    $REFL_CODE/Files/ssss/ssss_2023_spline_out.txt


    I do not think this is used anymore. It has been consolidated with the plot code.

    Parameters
    ----------
    iyear : int
        full year
    th : numpy array
        time values of some kind ... maybe fractional day of years?
    spline: fit, output of interpolate.BSpline
        needs doc
    delta_out : int
        how often you want the splinefit water level written, in seconds
    station : str
        station name
    txtdir : str
        output directory
    Hortho : float
        orthometric height used to convert RH to something more sea level like
        meters
    """

    if delta_out == 0:
        print('No spline values will be written because the interval is set to zero.')
    else:
        firstpoint  = float(th[0]); lastpoint = float(th[-1])
        s1 = math.floor(firstpoint) ; s2 = math.ceil(lastpoint)
        ndays = s2-s1 # number of days
        # how many values you want in the linspace world
        numvals = 1 + int(ndays*86400/delta_out)
        # this should be evenly defined fractional doy
        tplot = np.linspace(s1, s2, numvals,endpoint=True)
        # then fit it
        spline_even = spline(tplot)
        N = len(tplot)

        # but you only want those values when we have data ....
        splinefileout =  txtdir + '/' + station + '_' + str(iyear) + '_spline_out_orig.txt'
        print('Writing evenly sampled file to: ', splinefileout)
        vn = ' gnssrefl v' + str(g.version('gnssrefl'))
        fout = open(splinefileout,'w+')
        fout.write('{0:1s}  {1:30s}  \n'.format('%','station: ' + station + vn))
        fout.write('{0:1s}  {1:30s}  \n'.format('%','This is NOT observational data - be careful when interpreting it.'))
        fout.write('{0:1s}  {1:30s}  \n'.format('%','If the data are not well represented by the spline functions, you will '))
        fout.write('{0:1s}  {1:30s}  \n'.format('%','have a very poor representation of the data. I am also writing out station '))
        fout.write('{0:1s}  {1:30s}  {2:8.3f} \n'.format('%','orthometric height minus RH, where Hortho (m) is ', Hortho  ))
        fout.write('{0:1s}  {1:30s}  \n'.format('%','This assumes RH is measured relative to the L1 phase center.  '))
        fout.write('{0:1s}  {1:30s}  \n'.format('%','MJD             RH(m) YYYY MM DD  HH  MM  SS   quasi-sea-level(m)'))
        fout.write('{0:1s}  {1:30s}  \n'.format('%','(1)              (2)  (3) (4) (5) (6) (7) (8)    (9)'))


        dtime = False
        for i in range(0,N):
            modjul = g.fdoy2mjd(iyear,tplot[i])
            doy = math.floor(tplot[i])
            utc= 24*(tplot[i] - doy)
            bigt,yy,mm,dd,hh,mi,ss = g.ymd_hhmmss(iyear,doy,utc,dtime)
            if (tplot[i] >= firstpoint) & (tplot[i] <= lastpoint):
                fout.write('{0:15.7f}  {1:10.3f} {2:4.0f} {3:3.0f} {4:3.0f} {5:3.0f} {6:3.0f} {7:3.0f} {8:10.3f} \n'.format(
                    modjul, spline_even[i], yy,mm,dd,hh,mi,ss, Hortho-spline_even[i]))
        fout.close()

def testing_nvals(Gval, Rval, Eval, Cval):
    """
    writing the number of observations per constellation as a test. not currently used

    Parameters
    Gval: numpy array
        GPS RH values
    Rval : numpy array
        GLONASS RH values
    Eval : numpy array
        Galileo RH values
    Cval : numpy
        Beidou RH values

    writes to a file - kristine.txt returns nothing

    """
    fouting = open('kristine.txt','w+')
    for i in range(0,len(Gval)):
        fouting.write(' {0:3.0f} {1:4.0f} {2:4.0f} {3:4.0f} {4:4.0f} \n'.format(i, Gval[i], Rval[i], Eval[i], Cval[i]))
    fouting.close()

    return


def numsats_plot(station,tval,nval,Gval,Rval,Eval,Cval,txtdir,fs,hires_figs,year):
    """
    makes the plot that summarizes the number of satellites in each
    constellation per epoch

    Parameters
    ----------
    station : str
        name of the station
    tval : numpy array
        datetime objects?
    nval : numpy array
        number of total satellites at a given epoch
    Gval : numpy array
        number of galileo satellites at an epoch
    Rval : numpy array
        number of glonass satellites at an epoch
    Eval : numpy array
        number of galileo satellites at an epoch
    Cval : numpy array
        number of beidou satellites at an epoch
    txtdir : str
        where results are stored
    fs : int
        fontsize for the plots
    hires_figs : bool
        try to plot high resolution
    year : int
        calendar year

    """

    fig,ax=plt.subplots(figsize=(10,5))
    ax.plot(tval,nval,'ko',label='Total',markersize=3)
    if (np.sum(Gval) > 0):
        ax.plot(tval,Gval,'bo',label='GPS',markersize=3)
    if (np.sum(Rval) > 0):
        ax.plot(tval,Rval,'ro',label='GLO',markersize=3)
    if (np.sum(Eval) > 0):
        ax.plot(tval,Eval,'o',color='orange',label='GAL',markersize=3)
    if (np.sum(Cval) > 0):
        ax.plot(tval,Cval,'co',label='BEI',markersize=3)

    plt.legend(loc="best",fontsize=fs)
    #ax.legend(loc='best')
    plt.title(station + ': number of RH retrievals each day',fontsize=fs)
    plt.xticks(rotation =45,fontsize=fs); plt.yticks(fontsize=fs)
    plt.grid()

    # was trying to do this, but life is short
    #if hires_figs:
#        https://stackoverflow.com/questions/4700614/how-to-put-the-legend-outside-the-plot
    # Shrink current axis's height by 10% on the bottom
        #box = ax.get_position()

        #ax.set_position([box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9])
        #ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05), fancybox=True, shadow=True, ncol=5)

    # Put a legend below current axis
        #ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05), fancybox=True, shadow=True, ncol=4)
        #ax.set_position([box.x0, box.y0 + box.height * 0.5, box.width, box.height ])
        #ax.legend(loc='center left', bbox_to_anchor=(1, 0.3))


    fig.autofmt_xdate()
    plotname = txtdir + '/' + station + '_' + str(year) + '_Subnvals.png'
    if hires_figs:
        plotname = txtdir + '/' + station + '_' + str(year) +'_Subnvals.eps'

    plt.savefig(plotname,dpi=300)
    print('Plot file saved as: ', plotname)


def rh_plots(otimes,tv,station,txtdir,year,d1,d2,percent99):
    """
    overview plots for rh_plot

    Parameters
    ----------
    otimes : numpy array of datetime objects
        observation times
    tv : numpy array
        gnssrefl results written into this variable using loadtxt
    station : str
        station name, only used for the title
    txtdir : str
        directory where the plots will be written to
    year: int
        what year is being analyzed
    d1 : int
        minimum day of year
    d2 : int
        maximum day of year
    percent99 : bool
        whether you want only the 1-99 percentile plotted

    """
    if d1 == 1 and d2 == 366:
        # these are the defaults
        setlimits = False
    else:
        setlimits = True
        yyy,mm,dd = g.ydoy2ymd(year, d1)
        th1 = datetime.datetime(year=year, month=mm, day=dd)
        yyy,mm,dd = g.ydoy2ymd(year, d2)
        th2 = datetime.datetime(year=year, month=mm, day=dd)
    # this is not working, so just setting it to false, cause who cares!
    setlimits = False
    fs = 10
    fig,(ax2,ax3)=plt.subplots(2,1,sharex=True)
        # put some azimuth information on it
    colors = tv[:,5]
        # ax.plot( otimes, tv[:,2], '.')
        # https://matplotlib.org/stable/gallery/lines_bars_and_markers/scatter_with_legend.html
    # help from felipe nievinski
    #tmp = numpy.percentile(tv[:,2], [1 99])
    #matplotlib.pyplot.ylim(tmp[1], tmp[2])

    scatter = ax2.scatter(otimes,tv[:,2],marker='o', s=10, c=colors)
    colorbar = fig.colorbar(scatter, ax=ax2)
    colorbar.set_label('deg', fontsize=fs)
    ax2.set_ylabel('meters',fontsize=fs)
    ax2.set_title('Azimuth',fontsize=fs)
    #ax1.title.set_text('Azimuth')
    plt.xticks(rotation =45,fontsize=fs); plt.yticks(fontsize=fs)
    fig.suptitle( station.upper() + ' Reflector Heights', fontsize=fs)

    if setlimits:
        ax2.set_xlim((th1, th2))

    # now doing 1 and 99%
    p1 = 0.01; p2 = 0.99
    #lowv, highv = my_percentile(tv[:,2],p1, p2)
    # this crashed my docker build - but it turned out to have nothing to do with it
    yl = np.percentile(tv[:,2] ,[1 ,99])
    print('percentile values',yl)
    # I think this is a typo?
    #lowv = yl[0]; highv = y[1]
    lowv = yl[0]; highv = yl[1]
    
    if percent99:
        ax2.set_ylim((lowv,highv))

    ax2.invert_yaxis()
    ax2.grid(True)

    fig.autofmt_xdate()

#    put some amplitude information on it
    # https://matplotlib.org/stable/gallery/lines_bars_and_markers/scatter_with_legend.html
    colors = tv[:,6]

    scatter = ax3.scatter(otimes,tv[:,2],marker='o', s=10, c=colors)
    colorbar = fig.colorbar(scatter, ax=ax3)
    ax3.set_ylabel('meters',fontsize=fs)
    colorbar.set_label('v/v', fontsize=fs)
    plt.xticks(rotation =45,fontsize=fs); plt.yticks(fontsize=fs)
    ax3.set_title('Amplitude',fontsize=fs)
    ax3.grid(True)
    if setlimits:
        ax3.set_xlim((th1, th2))

    if percent99:
        ax3.set_ylim((lowv, highv ))
    ax3.invert_yaxis()
    fig.autofmt_xdate()

    if percent99:
        plotname = txtdir + '/' + station + '_rh2_99.png'
    else:
        plotname = txtdir + '/' + station + '_rh2.png'


    print('png file saved as: ', plotname)
    plt.savefig(plotname,dpi=300)


def two_stacked_plots(otimes,tv,station,txtdir,year,d1,d2,hires_figs):
    """
    This actually makes three stacked plots - not two, LOL
    It gives an overview for quality control

    Parameters
    ----------
    otimes : numpy array of datetime objects
        observations times
    tv : numpy array
        gnssrefl results written into this variable using loadtxt
    station : str
        station name, only used for the title
    txtdir : str
        where the plots will be written to
    year: int
        what year is being analyzed
    d1 : int
        minimum day of year
    d2 : int
        maximum day of year
    hires_figs : bool
        true for eps instead of png

    """
    if d1 == 1 and d2 == 366:
        # these are the defaults
        setlimits = False
    else:
        setlimits = True
        yyy,mm,dd = g.ydoy2ymd(year, d1)
        th1 = datetime.datetime(year=year, month=mm, day=dd)
        yyy,mm,dd = g.ydoy2ymd(year, d2)
        th2 = datetime.datetime(year=year, month=mm, day=dd)
    # this is not working, so just setting it to false, cause who cares!
    setlimits = False
    fs = 12
    fig,(ax1,ax2,ax3)=plt.subplots(3,1,sharex=True,figsize=(10,8))
    #fig,(ax1,ax2,ax3)=plt.subplots(3,1,sharex=True)
    i = (tv[:,10] < 100)
    colors = tv[:,10]
    scatter = ax1.scatter(otimes,tv[:,2],marker='o', s=10, c=colors)
    colorbar = fig.colorbar(scatter, ax=ax1)
    colorbar.set_label('Frequency', fontsize=fs)
    ax1.set_title('Signals',fontsize=fs)
    plt.xticks(rotation =45,fontsize=fs); 
    ax1.set_ylabel('RH (m)',fontsize=fs)
    plt.yticks(fontsize=fs)
    ax1.invert_yaxis()
    ax1.grid(True)
    #fig.suptitle( station.upper() + ' Reflector Heights', fontsize=fs)
    if setlimits:
        ax1.set_xlim((th1, th2))
    fig.autofmt_xdate()

    #fig,(ax1,ax2)=plt.subplots(2,1,sharex=True)
        # put some azimuth information on it
    colors = tv[:,5]
        # ax.plot( otimes, tv[:,2], '.')
        # https://matplotlib.org/stable/gallery/lines_bars_and_markers/scatter_with_legend.html
    scatter = ax2.scatter(otimes,tv[:,2],marker='o', s=10, c=colors)
    colorbar = fig.colorbar(scatter, ax=ax2)
    colorbar.set_label('deg', fontsize=fs)
    ax2.set_ylabel('RH (m)',fontsize=fs)
    ax2.set_title('Azimuth',fontsize=fs)
    #ax1.title.set_text('Azimuth')
    plt.xticks(rotation =45,fontsize=fs); plt.yticks(fontsize=fs)
    ax2.invert_yaxis()
    ax2.grid(True)
    if setlimits:
        ax2.set_xlim((th1, th2))
    fig.autofmt_xdate()

# put some amplitude information on it
    colors = tv[:,6]
    # ax.plot( otimes, tv[:,2], '.')
    # https://matplotlib.org/stable/gallery/lines_bars_and_markers/scatter_with_legend.html
    scatter = ax3.scatter(otimes,tv[:,2],marker='o', s=10, c=colors)
    colorbar = fig.colorbar(scatter, ax=ax3)
    ax3.set_ylabel('RH (m)',fontsize=fs)
    colorbar.set_label('v/v', fontsize=fs)
    plt.xticks(rotation =45,fontsize=fs); plt.yticks(fontsize=fs)
    ax3.set_title('Amplitude',fontsize=fs)
    ax3.invert_yaxis()
    ax3.grid(True)
    if setlimits:
        ax3.set_xlim((th1, th2))
    fig.autofmt_xdate()

    plotname = txtdir + '/' + station + '_' + str(year) + '_combined.png'
    if hires_figs:
        plotname = txtdir + '/' + station + '_' + str(year) + '_combined.eps'
        plt.savefig(plotname,dpi=300)
    else:
        plt.savefig(plotname,dpi=300)
    print('Plot file saved as: ', plotname)

def stack_two_more(otimes,tv,ii,jj,stats, station, txtdir, sigma,kplt,hires_figs,year):
    """
    makes a plot of the reflector heights before and after minimal editing

    Parameters
    ----------
    otimes : numpy array of datetime objects 
        observation times
    tv : numpy array
        variable with the gnssrefl LSP results
    ii : numpy array
        indices of good data
    jj : numpy array
        indices of bad data
    station : str
        station name
    txtdir : str
        directory where plots will be written
    sigma : float
        what kind of standard deviation is used for outliers (3sigma, 2.5 sigma etc)
    kplt : bool
        make extra plot for kristine
    year : int 
        calendar year
    """
    fs = 10
    otimesarray = np.asarray(otimes)
    # new plot 
    plt.figure(figsize=(10,5))
    plt.plot(tv[:,5],tv[:,2], '.',markersize=4,label='obs')
    plt.plot(tv[ii,5],tv[ii,2], 'ro',markersize=4,label='outliers')
    plt.xlabel('Azimuth (degrees)')
    plt.ylabel('Reflector Height (m)')
    plt.title('Quick Plot of RH with respect to Azimuth')
    plt.gca().invert_yaxis()
    plt.legend(loc="best")
    plt.grid()
    plotname_stem = txtdir + '/' + station + '_' + str(year) + '_outliers_wrt_az'
    if hires_figs:
        plotname = plotname_stem + '.eps'
    else:
        plotname = plotname_stem + '.png'

    plt.savefig(plotname,dpi=300)
    print('Plot file saved as: ', plotname)

    #    fig=plt.figure(figsize=(10,6))
    fig = plt.figure(figsize=(10,6))
    colors = tv[:,5]
# put some amplitude information on it
# ax.plot( otimes, tv[:,2], '.')
# https://matplotlib.org/stable/gallery/lines_bars_and_markers/scatter_with_legend.html


    ax1 = fig.add_subplot(211)
    plt.plot(otimes,tv[:,2], '.',color='gray',label='arcs')
    plt.plot(stats[:,0], stats[:,1], '.',markersize=4,color='blue',label='daily avg')
    slabel = str(sigma) + ' sigma'
    plt.plot(stats[:,0], stats[:,1]-sigma*stats[:,2], '--',color='black',label=slabel)
    plt.plot(stats[:,0], stats[:,1]+sigma*stats[:,2], '--',color='black')
    plt.plot(otimesarray[ii],tv[ii,2], 'r.',markersize=4,label='outliers')
    #plt.plot(otimesarray[ii],tv[ii,2], '.',color='red',label='outliers',markersize=12)
    plt.legend(loc="best",bbox_to_anchor=(0.95, 0.9),prop={"size":fs-2})
    plt.title('Raw ' + station.upper() + ' Reflector Heights', fontsize=fs)
    plt.ylabel('meters',fontsize=fs)
    plt.gca().invert_yaxis()
    plt.xticks(rotation =45,fontsize=fs); plt.yticks(fontsize=fs)
    plt.grid() ; fig.autofmt_xdate()
    # get the limits so you can use thme on the next plot
    #aaa, bbb = plt.ylim()
    savey1,savey2 = plt.ylim()  

    ax2 = fig.add_subplot(212)
    plt.plot(otimesarray[jj],tv[jj,2], '.',color='green',label='arcs')
    plt.gca().invert_yaxis()
    plt.ylabel('meters',fontsize=fs)
    plt.xticks(rotation =45,fontsize=fs); plt.yticks(fontsize=fs)
    plt.title('Edited ' + station.upper() + ' Reflector Heights', fontsize=fs)

    plt.grid() ; fig.autofmt_xdate()
    plotstem = txtdir + '/' + station + '_' + str(year) + '_outliers'
    if hires_figs:
        plotname = plotstem + '.eps'
        plt.savefig(plotname,dpi=300)
    else:
        plotname = plotstem + '.png'
        plt.savefig(plotname,dpi=300)

    plt.ylim((savey1, savey2))
    print('Plot file saved as: ', plotname)
    if kplt:
        fig = plt.figure()
        ax1 = fig.add_subplot(211)
        ddd = - (tv[jj,2] - np.max(tv[jj,2]))
        plt.plot(otimesarray[jj],ddd, '.',color='blue')
        plt.title('Relative Sea Level Measured with Reflected GNSS Signals:' + station.upper(), fontsize=fs)
        plt.ylabel('meters',fontsize=fs)
        plt.xticks(rotation =45,fontsize=fs-1); plt.yticks(fontsize=fs-1)
        plt.ylim((-0.1, 1.1*np.max(ddd)))
        plt.suptitle(f"St Michael, Alaska ", size=12)
        plt.grid()

def writeout_spline_outliers(tvd_bad,txtdir,residual,filename):
    """

    Write splinefit outliers to a text file. 

    Parameters
    ----------
    tvd_bad : numpy array
        output of the lomb scargle calculations
    txtdir : str
        directory for the output, i.e. $REFL_CODE/FiLes/station
    residual : numpy array
        RH outliers in units of meters (!)
    filename : str
        name of file being written

    """
    nr,nc=tvd_bad.shape
    if nr > 0:
        f = txtdir + '/' + filename
        print(nr, ' Outliers written to: ', f)
        fout = open(f, 'w+')
        # put in a header
        fout.write('{0:3s} sat azim deltT-min outlier-m fracDOY MJD  OrigRH meanE PkNoise \n'.format('%'))
        fout.write('{0:3s} (1) (2)   (3)        (4)      (5)    (6)   (7)    (8)   (9)\n'.format('%'))
        for w in range(0,nr):
            fy = tvd_bad[w,1] + tvd_bad[w,4]/24 # fractional day of year
            deltaT = tvd_bad[w,14]
            mjd = tvd_bad[w,15]
            # average elevation angle
            elAv = 0.5*(tvd_bad[w,8] + tvd_bad[w,7])
            fout.write('{0:3.0f} {1:7.2f} {2:7.2f} {3:7.2f} {4:9.3f} {5:15.7f} {6:7.2f} {7:7.2f} {8:5.2f}\n'.format( 
                tvd_bad[w,3], tvd_bad[w,5], deltaT,residual[w],fy,mjd,tvd_bad[w,2],elAv,tvd_bad[w,13]))
        fout.close()

    return

def print_badpoints(t,outliersize,txtdir,real_residuals):
    """
    prints outliers to a file.

    Parameters
    ----------
    t : numpy array
        lomb scargle result array of "bad points". Format given below
    outliersize : float
        outlier criterion, in meters
    txtdir : str
        directory where file is written
    real_residuals : numpy array of floats
        assume this is RH residuals in meters

    Returns
    -------
    writes to a file called outliers.txt in the Files/station area

    """
# format of t 

# (1)  (2)   (3) (4)  (5)     (6)   (7)    (8)    (9)   (10)  (11) (12) (13)    (14)     (15)    (16) (17) (18,19,20,21,22)
# year, doy, RH, sat,UTCtime, Azim, Amp,  eminO, emaxO,NumbOf,freq,rise,EdotF, PkNoise  DelT     MJD  refr  MM DD HH MM SS
# (0)  (1)   (2) (3)  (4)     (5)   6 )    (7)    (8)   (9)  (10) (11) (12)    (13)     (14)    (15)  (16) ... 

    m,n = t.shape
    f = txtdir + '/outliers.txt'
    if (m > 0):
        print(m, ' Outliers written to file: ', f)        
        fout = open(f, 'w+')
        for i in range(0,m):
            fout.write('doy {0:3.0f} sat {1:3.0f} azim {2:6.2f} fr {3:3.0f} pk2noise {4:5.1f} residual {5:5.2f} \n'.format( 
                t[i,1], t[i,3],t[i,5], t[i,10], t[i,13], real_residuals[i] ))
        fout.close()
    else:
        print('no outlier points to write to a file')

def writejsonfile(ntv,station, outfile):
    """
    subdaily RH values written out in json format

    This does not appear to be used

    Parameters
    -----------
    ntv : numpy of floats
        LSP results

    station : str
        4 ch station name

    outfile : str
        filename for output

    """
    print('You picked the json output')
    # dictionary
    #o = {}
    N= len(ntv)

    # year is in first column
    year  =  ntv[:,0].tolist()
    year =[str(int(year[i])) for i in range(N)];

    # day of year
    doy =  ntv[:,1].tolist()
    doy=[str(int(doy[i])) for i in range(N)];

    # UTC hour
    UTChour = ntv[:,4].tolist()
    UTChour = [str(UTChour[i]) for i in range(N)];

    # converted to ???
    timestamp = [quickTr(ntv[i,0], ntv[i,1], ntv[i,4]) for i in range(N)]

    # reflector height (meters)
    rh = ntv[:,2].tolist()
    rh=[str(rh[i]) for i in range(N)];

    # satellite number
    sat  = ntv[:,3].tolist()
    sat =[int(sat[i]) for i in range(N)];

    # frequency
    freq  = ntv[:,8].tolist()
    freq =[int(freq[i]) for i in range(N)];

    # amplitude of main periodogram (LSP)
    ampl  = ntv[:,6].tolist()
    ampl =[str(ampl[i]) for i in range(N)];

    # azimuth in degrees
    azim  = ntv[:,5].tolist()
    azim =[str(azim[i]) for i in range(N)];

    # edotF in units ??
    edotf  = ntv[:,9].tolist()
    edotf =[str(edotf[i]) for i in range(N)];

    # modified julian day
    #mjd = ntv[:,15].tolist()
    mjd = ntv[:,14].tolist()
    mjd=[str(mjd[i]) for i in range(N)];

    #column_names = ['timestamp','rh','sat','freq','ampl','azim','edotf','mjd']
    # now attempt to zip them
    l = zip(timestamp,rh,sat,freq,ampl,azim,edotf,mjd)
    dzip = [dict(zip(column_names, next(l))) for i in range(N)]
    # make a dictionary with metadata and data
    o={}
    # not setting lat and lon for now
    lat = "0"; lon = "0";
    firstline = {'name': station, 'latitude': lat, 'longitude': lon}
    o['metadata'] = firstline
    o['data'] = dzip
    outf = outfile
    with open(outf,'w+') as outf:
        json.dump(o,outf,indent=4)

    return True

def rhdot_plots(th,correction,rhdot_at_th, tvel,yvel,fs,station,txtdir,hires_figs,year):
    """
    makes the rhdot correction plots

    Parameters
    ----------
    th : numpy array
        time of obs, MJD
    correction : numpy array
        rhcorrections in meters
    rhdot_at_th : numpy array of floats
        spline fit for rhdot in meters
    tvel : numpy array of floats
        time for surface velocity in MJD
    yvel : numpy array of floats
        surface velocity in m/hr
    fs : integer
        fontsize 
    station : str
        station name
    txtdir : str
        file directory for output
    hires_figs : bool
        whether you want eps instead of png 
    year : int
        calendar year

    """
    mjd0 = g.fdoy2mjd(year, th[0] ) # MJD of first point
    # not being used


    fig=plt.figure(figsize=(10,6))
    plt.subplot(2,1,1)
    th_obs = mjd_to_obstimes(th)

    plt.plot(th_obs, correction,'b.')

    plt.ylabel('meters',fontsize=fs);
    #plt.xlim((np.min(th), np.max(th)))
    plt.grid()
    plt.title('The RH correction from RHdot effect ')
    #plt.xlim((th_obs[0], th_obs[-1])) #fig.autofmt_xdate()

    plt.subplot(2,1,2)
    #mjd1 = g.fdoy2mjd(year, tvel[0] ) # MJD of first point
    #tvel_obs = mjd_to_obstimes(tvel+mjd1)

    # ???
    A1 = np.min(th)  ; A2 = np.max(th) 
    jj = (tvel >= A1) & (tvel <= A2)

    plt.plot(th_obs, rhdot_at_th,'b.',label='at GNSS obs')
    # disaster - did not work
    #plt.plot(tvel_obs, yvel,'c-', label='spline fit')

    tvel_obstime = mjd_to_obstimes(tvel[jj])
    plt.plot(tvel_obstime, yvel[jj],'c-', label='spline fit')
    #plt.plot(tvel[jj], yvel[jj],'c-', label='spline fit')

    plt.legend(loc="upper left")
    plt.grid()
    plt.title('surface velocity')
    plt.ylabel('meters/hour'); 
    #plt.xlabel('days of the year') 
    fig.autofmt_xdate()

    #plt.xlim((A1,A2))
    #plt.xlim((th[0], th[-1])) #fig.autofmt_xdate()

    if hires_figs:
        g.save_plot(txtdir + '/' + station + '_rhdot3.eps')
    else:
        g.save_plot(txtdir + '/' + station + '_rhdot3.png')


def find_ortho_height(station,extension):
    """
    Find orthometric (sea level) height used in final subdaily spline output 
    and plots. This value should be defined for the GPS L1 phase center of 
    the GNSS antenna as this is what is assumed in the subdaily code.

    Parameters
    ----------
    station : str
        4 ch station name
    extension : str
        gnssir analysis, extension mode

    Returns
    -------
    Hortho : float
        orthometric height from gnssir json analysis file as 
        defined as Hortho, in meters. If your preferred value 
        for Hortho is not present, it is calculated from the  
        ellipsoidal height and EGM96. 

    """

    lsp = guts2.read_json_file(station, extension)

    if 'Hortho' in lsp:
        # found in the lsp file
        Hortho = lsp['Hortho']
    else:
        # calculate from EGM96
        geoidC = g.geoidCorrection(lsp['lat'],lsp['lon'])
        Hortho = lsp['ht']- geoidC

    return Hortho

def mirror_plot(tnew,ynew,spl_x,spl_y,txtdir,station,beginT,endT):
    """
    Makes a plot of the spline fit to the mirrored RH data
    Plot is saved to txtdir as  station_rhdot1.png

    Parameters
    ----------
    tnew : numpy of floats
        time in MJD
    ynew : numpy of floats
        RH in meters 
    spl_x : numpy of floats
        time in days of year
    spl_y : numpy of floats
        smooth RH, meters
    txtdir : str
        directory for plot
    station : str
        name of station for title
    beginT : float
        first measurement time (MJD) for RH measurement
    endT : float
        last measurement time (MJD) for RH measurement

    """
    fig=plt.figure(figsize=(10,4))
    i = (spl_x > beginT) & (spl_x < endT) # to avoid wonky points in spline
    #if (beginT > 380):
    if True:
        objtime = mjd_to_obstimes(tnew)
        plt.plot(objtime,ynew, 'b.', label='obs+fake obs')
        plt.plot(mjd_to_obstimes(spl_x),spl_y,'-', color='orange',label='spline all')
        plt.plot(mjd_to_obstimes(spl_x[i]),spl_y[i],'c-', label='spline obs')
        plt.xlabel('MJD')
    #else:
    #    plt.plot(tnew,ynew, 'b.', label='obs+fake obs')
    #    plt.plot(spl_x,spl_y,'o-', label='spline for all')
    #    plt.plot(spl_x[i],spl_y[i],'c-', label='spline for obs')
    #    plt.xlabel('days of the year')

    plt.title('Mirrored obs and spline fit ')
    plt.legend(loc="upper left")
    plt.ylabel('meters')
    #plt.xlim((beginT, endT))

    plt.grid()
    fig.autofmt_xdate()

    g.save_plot(txtdir + '/' + station + '_rhdot1.png')
    # plot to the screen ... 
    #plt.close()

def quickTr(year, doy,frachours):
    """
    takes timing from lomb scargle code (year, doy) and UTC hour (fractional)
    and returns a date string

    Parameters
    ----------
    year : int
        full year
    doy : int
        day of year
    frachours : float
        real-valued UTC hour 

    Returns
    -------
    datestring : str
         date ala YYYY-MM-DD HH:MM:SS
    """
    year = int(year); doy = int(doy); frachours = float(frachours)
    # convert doy to get month and day
    d = datetime.datetime(year, 1, 1) + datetime.timedelta(days=(doy-1))
    month = int(d.month)
    day = int(d.day)

    hours = int(np.floor(frachours))
    leftover = 60*(frachours - hours)
    minutes = int(np.floor(leftover))
    leftover_hours  = frachours - (hours + minutes/60)
    seconds = int(leftover_hours*3600)
    #print(frachours, hours,minutes,leftover_seconds)

    jd = datetime.datetime(year,month, day,hours,minutes,seconds)
    datestring = jd.strftime("%Y-%m-%d %H:%M:%S")


    return datestring


def subdaily_resids_last_stage(station, year, th, biasCor_rh, spline_at_GPS, fs, strsig, hires_figs, 
                               txtdir, ii, jj, th_even,spline_whole_time):
    """
    Makes the final residual plot for subdaily (after RHdot and IF correction made).
    Returns the bad points ...


    Allows either the original or multiyear option..


    Parameters
    ----------
    station : str
        4 character station name
    year : int
        calendar year
    th : numpy array of ??
        time variable of some kind, fractional day of year ?
    biasCor_rh : numpy array of floats
        refl hgts that have been corrected for RHdot and IF
    spline_at_GPS : numpy array of floats
        RH derived From the spline fit and calculated at GPS time tags
    fs : int
        font size
    strsig : str
        sigma string to go on the legend
    hires_figs : bool
        whether to save the plots with better resolution
    txtdir : str
        directory where the plot will be saved
    ii : numpy array
        indices of the outliers?
    jj : numpy array
        indices of the values to keep?
    th_even : numpy array
        evenly spaced time values, day of year
    spline_whole_time : numpy array of flots
        splinefit for ???

    Returns
    -------
    badpoints2 : numpy array of floats
         RH residuals
    """
    # this means you are already in MJD ... 
    #if (th[0] > 400):
    if True:
        th_obs = mjd_to_obstimes(th)
        th_even_obs = mjd_to_obstimes(th_even)
    #else:
    #    # convert obs to mjd and then obstimes
    #    mjd0 = g.fdoy2mjd(year, th[0])
    #    th_obs = mjd_to_obstimes(mjd0 + th - th[0])
    # convert the spline to mjd and then obstimes
    #    mjd1 = g.fdoy2mjd(year, th_even[0])
    #    th_even_obs = mjd_to_obstimes(mjd1 + th_even - th_even[0])


    # these are now plotted in datetime via mjd translation
    fig=plt.figure(figsize=(10,5))
    plt.plot(th_obs, biasCor_rh, 'b.', label='RH ' + strsig)
    plt.plot(th_even_obs, spline_whole_time, '-',color='orange',label='spline')
    plt.plot(th_obs[ii], biasCor_rh[ii], 'rx', label='outliers')

    plt.legend(loc="best",fontsize=fs)
    plt.grid() ; plt.gca().invert_yaxis()
    plt.ylabel('meters',fontsize=fs); 
    plt.title(station.upper() + ' RH with RHdot and InterFrequency Corrections Applied',fontsize=fs)
    fig.autofmt_xdate()

    # put hires_figs boolean here
    if hires_figs:
        g.save_plot(txtdir + '/' + station + '_rhdot4.eps')
    else:
        g.save_plot(txtdir + '/' + station + '_rhdot4.png')

    # this figure is not sent to the screen
    fig=plt.figure(figsize=(10,5))
    plt.plot(th_obs, biasCor_rh - spline_at_GPS, 'b.',label='all residuals')
    plt.title('Station:' + station + ' Residuals to new spline fit',fontsize=fs)
    plt.grid()
    plt.ylabel('meters',fontsize=fs)

    plt.plot(th_obs[ii], (biasCor_rh -spline_at_GPS)[ii], 'r.',label='outliers')
    # will write these residauls out to a file
    badpoints2 =  (biasCor_rh -spline_at_GPS)[ii]
    plt.legend(loc="upper left",fontsize=fs)
    fig.autofmt_xdate()

    # print('\n RMS with frequency biases and RHdot taken out (m) ', np.round(newsigma,3) , '\n' )
    if hires_figs:
        g.save_plot(txtdir + '/' + station + '_rhdot5.eps')
    else:
        g.save_plot(txtdir + '/' + station + '_rhdot5.png')

    plt.close() # 

    return badpoints2

def RH_ortho_plot2( station, H0, year,  txtdir, fs, time_rh, rh, gap_min_val,th,spline,delta_out,csvfile,gap_flag):
    """

    Makes a plot of the final spline fit to the data. Output time interval controlled by the user.

    It also now writes out the file with the spline fit

    Parameters
    ----------
    station : str
        name of station, 4 ch
    H0 : float
        datum correction (orthometric height) to convert RH to MSL data, in meters
    year : int
        year of the time series (ultimately should not be needed)
    txtdir : str
       location of plot
    fs : int
        fontsize
    time_rh : numpy of floats
        time of rh values, in fractional doy I believe
    rh : numpy of floats
        refl hgt in meters
    gap_min_val : float
        minimum length gap allowed, in day of year units
    th : numpy floats
        time values in MJD
    spline : output of interpolate.BSpline
        used for fitting 
    delta_out : int
        how often spline is printed, in seconds
    csvfile : bool
        print out csv instead of plain txt
    gap_flag : bool
        whether to write 999 in file where there are gaps
    """

    #print('Entering RH_ortho_plot2')
    firstpoint = float(th[0]); lastpoint = float(th[-1])
    s1 = math.floor(firstpoint); s2 = math.ceil(lastpoint)
    ndays = s2-s1 # number of days
    numvals = 1 + int(ndays*86400/delta_out)
    tp=np.linspace(s1,s2,numvals,endpoint= True)
    #print('RH_ortho_plot2, s1, s2, numvals,ndays')
    #print(s1, s2, numvals,ndays)

    # previous tp definition goes from beginning of the first day to the end of last day - but 
    # does not recognize that the file could have started well beyond that first point 
    # and it might end say 8 hours into the last day
    ii = (tp >= firstpoint) & (tp <= lastpoint)
    tp = tp[ii]

    # this means it is already in MJD ???
    #if (th[0] > 400): 

    if True: 
        # don't think this boolean is used anymore.  it is not
        # truly multiyear, but more a flag that you are using MJD
        multiyear = True
        mjd_new = tp
        #print('multiyear is true, start at ', mjd_new[0])
        mjd_new_obstimes = mjd_to_obstimes(mjd_new)
        spline_new = spline(tp)
    #else:
    #    multiyear = False
    #    mjd1 = g.fdoy2mjd(year, tp[0] ) # 
    #    #print(year, tp[0], mjd1,np.floor(mjd1))
    #    mjd_new = np.floor(mjd1) + (tp - tp[0])
    #    print('multiyear is false, start at ', mjd_new[0])
    #    mjd_new_obstimes = mjd_to_obstimes(mjd_new)
    #    spline_new = spline(tp)

    N_new = len(mjd_new_obstimes)

    # looks like I identified the gaps in day of year units - 
    # but then did the implementation in mjd and then datetime ...
    #splinefileout =  txtdir + '/' + station + '_' + str(year) + '_spline_out.txt'
    if csvfile:
        splinefileout =  txtdir + '/' + station +  '_spline_out.csv'
    else:
        splinefileout =  txtdir + '/' + station +  '_spline_out.txt'


    print('Writing evenly sampled file to: ', splinefileout)
    fout = open(splinefileout,'w+')
    vn = station + ' gnssrefl v' + str(g.version('gnssrefl'))
    fout.write('{0:1s}  {1:30s}  \n'.format('%','station ' + vn))
    fout.write('{0:1s}  {1:60s}  \n'.format('%','TIME TAGS ARE IN UTC/999 values mean there was a large gap and no spline value is available.'))
    fout.write('{0:1s}  {1:30s}  \n'.format('%','This is NOT observational data - be careful when interpreting it.'))
    fout.write('{0:1s}  {1:30s}  \n'.format('%','If the data are not well represented by the spline functions, you will '))
    fout.write('{0:1s}  {1:30s}  \n'.format('%','have a very poor representation of the data. I am also writing out station '))
    fout.write('{0:1s}  {1:30s}  {2:8.3f} \n'.format('%','orthometric height minus RH, where Hortho (m) is ', H0  ))
    #fout.write('{0:1s}  {1:30s}  \n'.format('%','This assumes RH is measured relative to the L1 phase center.  '))
    #fout.write('{0:1s}  {1:30s}  \n'.format('%','MJD, RH(m), YY,MM,DD,HH,MM,SS, quasi-sea-level(m)'))
    if csvfile:
        fout.write('{0:1s}  {1:30s}  \n'.format('%','MJD,             RH(m), YYYY, MM, DD, HH, MM, SS,  quasi-sea-level(m)'))
        fout.write('{0:1s}  {1:30s}  \n'.format('%','(1),              (2),  (3), (4),(5), (6),(7),(8),   (9)'))
    else:
        fout.write('{0:1s}  {1:30s}  \n'.format('%','MJD              RH(m)  YYYY  MM  DD  HH  MM  SS   quasi-sea-level(m)'))
        fout.write('{0:1s}  {1:30s}  \n'.format('%','(1)               (2)   (3)  (4) (5)  (6) (7) (8)    (9)'))


    # difference function to find time between all rh measurements
    gdiff = np.diff( time_rh )
    if multiyear:
        mjdt = time_rh
        mjd = mjdt[0:-1]
    else:
        mjdx = g.fdoy2mjd(year, time_rh[0] ) # 
        mjdt = mjdx + (time_rh - time_rh[0])
        mjd = mjdt[0:-1]


    # find indices of gaps  that are larger than a certain value
    ii = (gdiff > gap_min_val)
    #print('gap_min_val',gap_min_val, gdiff)
    N = len(mjd[ii])
    Ngdiff = len(gdiff)

    dlist = mjd_new_obstimes.tolist()


    # i have not figured this one out yet ... the horror
    #if not multiyear:
    if True:
    # get the integer values to write out  to the text file ...
        theyear = [y.year for y in dlist ]
        xm = [y.month for y in dlist]; xd = [y.day for y in dlist]
        xh = [y.hour for y in dlist]; xmin = [y.minute for y in dlist]
        xs = [y.second for y in dlist]

        if (Ngdiff > 0):
            for i in range(0,Ngdiff):
                if ii[i]:
                    try:
                        e0 = mjd[i] ; e1 = mjd[i+1] # beginning and ending of the gap
                    except:
                        # alerted to this problem by Felipe Nievinski.  
                        print('>>>>  DUNNO, some problem with the spline - maybe gap at end of day',mjd[i])
                        continue
                    bad_indices = (mjd_new > e0) & (mjd_new < e1 )
                    spline_new[bad_indices] = np.nan
                    mjd_new_obstimes[bad_indices] = np.datetime64("NaT")

    # write the spline values to a file, with gaps removed
        for i in range(0,N_new):
            if not np.isnan(spline_new[i]):
                rhout = spline_new[i]
                if csvfile:
                    fout.write('{0:15.7f}, {1:10.3f},{2:4.0f},{3:3.0f},{4:3.0f},{5:3.0f},{6:3.0f},{7:3.0f},{8:10.3f} \n'.format( mjd_new[i], rhout, theyear[i], xm[i], xd[i], xh[i], xmin[i], xs[i], H0-rhout))
                else:
                    fout.write('{0:15.7f}  {1:10.3f} {2:4.0f} {3:3.0f} {4:3.0f} {5:3.0f} {6:3.0f} {7:3.0f} {8:10.3f} \n'.format(mjd_new[i], rhout, theyear[i], xm[i], xd[i], xh[i], xmin[i], xs[i], H0-rhout))
            else:
                if gap_flag: 
                    rhout = spline_new[i]
                    if csvfile:
                        fout.write('{0:15.7f}, {1:10.3f},{2:4.0f},{3:3.0f},{4:3.0f},{5:3.0f},{6:3.0f},{7:3.0f},{8:10.3f} \n'.format( mjd_new[i], 999, theyear[i], xm[i], xd[i], xh[i], xmin[i], xs[i], 999))
                    else:
                        fout.write('{0:15.7f}  {1:10.3f} {2:4.0f} {3:3.0f} {4:3.0f} {5:3.0f} {6:3.0f} {7:3.0f} {8:10.3f} \n'.format(mjd_new[i], 999, theyear[i], xm[i], xd[i], xh[i], xmin[i], xs[i], 999))


    fout.close()

    fig=plt.figure(figsize=(10,5))
    plt.plot(mjd_new_obstimes, H0 -spline_new, 'b-',linewidth=2)
    plt.grid()
    plt.ylabel('meters',fontsize=fs)
    plt.title(station.upper() + ' Water Level from GNSS-IR', fontsize=fs)
    fig.autofmt_xdate()

    pfile = txtdir + '/' + station + '_H0.png'
    g.save_plot(pfile)

    return

def pickup_subdaily_json_defaults(xdir, station, extension):
    """
    picks up an existing gnssir analysis json. augments
    with subdaily parameters if needed. Returns the dictionary.

    Parameters
    ----------
    xdir : str
        REFL_CODE code location
    station : str
        name of station
    extension : str
        possible extension location

    Returns
    -------
    lsp : dictionary 
        contents of gnssir json
    """

    # changed default in subdaily to None instead of '' So have to check
    if extension is None:
        lsp = guts2.read_json_file(station, '',noexit=True)
    else:
        lsp = guts2.read_json_file(station, extension,noexit=True)

    # check json for subdaily settings
    # if values not found, then set them to None

    if 'subdaily_ampl' not in lsp:
        lsp['subdaily_ampl'] = None

    if 'subdaily_spline_outlier1' not in lsp:
        lsp['subdaily_spline_outlier1'] = None

    if 'subdaily_spline_outlier2' not in lsp:
        lsp['subdaily_spline_outlier2'] = None

    if 'subdaily_knots' not in lsp:
        lsp['subdaily_knots'] = None

    if 'subdaily_delta_out' not in lsp:
        lsp['subdaily_delta_out'] = None

    if 'subdaily_alt_sigma' not in lsp:
        lsp['subdaily_alt_sigma'] = None

    if 'subdaily_sigma' not in lsp:
        lsp['subdaily_sigma'] = None

    if 'subdaily_subdir' not in lsp:
        lsp['subdaily_subdir'] = None

    return lsp

def flipit3(tvd,col):
    """
    Third version of the flipit code. It takes a time series of RH values, extracts 
    24 hours of observations from the beginning and end of the series, uses
    them as fake data to make the spline fit stable.  
    Also fill the temporal gaps with fake data
    Previous versions assumed you were going to have full days of data at the beginning
    and end of the series.

    This version uses MJD rather than day of year for x-axis

    Parameters
    ----------
    tvd : numpy array of floats
        output of LSP runs. 
    col : integer
        column number (in normal speak) of the RH results
        in python-speak, this will have one subtracted from it

    Returns
    -------
    tnew : numpy array of floats
        time in MJD
    ynew : numpy array
        RH in meters 

    """
    nr,nc = np.shape(tvd)
    #print(nr,nc)
    # sort it just to make sure ...
    #tnew = tvd[:,1] + tvd[:,4]/24
    # use MJD
    tnew = tvd[:,15]
    # change from normal columns to python columns
    ynew = tvd[:,col-1]

    # these are in MJD 
    day0= tnew[0] # first observation time
    dayN = tnew[-1] # last observation time

    # these are the times relative to time zero
    middle = tnew-day0

    # use the first day
    ii = tnew < (day0+1)
    leftTime = -(tnew[ii]-day0)
    leftY = ynew[ii]

    # now use the last day. no idea if this will work
    ii = tnew > (dayN-1)
    rightY = np.flip(ynew[ii])
    rightTime = tnew[ii] -day0 + 1 

    tmp= np.hstack((leftTime,middle)) ; 
    th = np.hstack((tmp,rightTime))

    tmp = np.hstack((leftY, ynew )) ; 
    h = np.hstack((tmp, rightY))

    # and sort it ...
    ii = np.argsort(th)
    th = th[ii] ; h = h[ii]

    th = th + day0 # add day0 back in

    # now fill the gaps ... 
    fillgap = 1/24 # one hour fake values
    # ???
    gap = 5/24 # up to five hour gap allowed before warning

    tnew =[] ; ynew =[]; faket = [];
    # fill in gaps using variables called tnew and ynew
    Ngaps = 0
    for i in range(1,len(th)):
        d= th[i]-th[i-1] # delta in time in units of days ?
        if (d > gap):
            x0 = th[i-1:i+1] ; h0 = h[i-1:i+1]

            # only print out the gap information the first time thru
            if col == 3:
                print('Gap on MJD:', int(np.floor(x0[0])), ' lasting ', round(d*24,2), ' hours ')
            #print(d,x0,h0)
            Ngaps = Ngaps + 1
            f = scipy.interpolate.interp1d(x0,h0)
            # so this is fake data
            ttnew = np.arange(th[i-1]+fillgap, th[i], fillgap)
            yynew = f(ttnew)
            faket = np.append(faket, ttnew)
            # now append it to your real data
            tnew = np.append(tnew,ttnew)
            ynew = np.append(ynew,yynew)
        else:
            tnew = np.append(tnew,th[i])
            ynew = np.append(ynew,h[i])


    if (Ngaps > 3) and (col == 3):
        print('\nThis is a beta version of the rhdot/spline fit code - and does not')
        print('work well with gaps. You have been warned!\n')

    # sort again
    ii = np.argsort( tnew) 
    tnew = tnew[ii]
    ynew = ynew[ii]

    return tnew, ynew



import argparse
import numpy as np
import os
import sys
import subprocess
import matplotlib.pyplot as myplt


# my code
import gnssrefl.gps as g
import gnssrefl.subdaily as t

from gnssrefl.utils import str2bool


def parse_arguments():
    # must input start and end year
    parser = argparse.ArgumentParser()
    parser.add_argument("station", help="station name", type=str)
    parser.add_argument("year", default=None, type=int, help="for now one year at a time")
    parser.add_argument("-csvfile", default=None, type=str, help="set to True if you prefer csv to plain txt")
    parser.add_argument("-plt", default=None, type=str, help="set to False to suppress plots")
    parser.add_argument("-extension", default=None, type=str, help="solution subdirectory")
    parser.add_argument("-doy1", default=None, type=int, help="initial day of year")
    parser.add_argument("-doy2", default=None, type=int, help="end day of year")
    parser.add_argument("-ampl", default=None, type=float, help="new amplitude constraint")
    parser.add_argument("-azim1", default=None, type=int, help="new min azimuth")
    parser.add_argument("-azim2", default=None, type=int, help="new max azimuth")
    parser.add_argument("-h1", default=None, type=float, help="min RH (m)")
    parser.add_argument("-h2", default=None, type=float, help="max RH (m)")
    parser.add_argument("-peak2noise", default=None, type=float, help="new peak2noise constraint")

    args = parser.parse_args().__dict__

    # convert all expected boolean inputs from strings to booleans
    boolean_args = ['csvfile', 'plt' ]
    args = str2bool(args, boolean_args)

    # only return a dictionary of arguments that were added from the user - all other defaults will be set in code below
    return {key: value for key, value in args.items() if value is not None}


def rh_plot(station: str, year: int, csvfile: bool = False, plt: bool = True,
             extension: str = '', doy1: int = 1, doy2: int = 366, ampl: float = 0, 
             h1: float=0.0, h2: float=300.0, azim1: int=0, azim2: int = 360, 
             peak2noise: float = 0 ) :
    """
    Parameters
    ----------

    station : string
        4 character id of the station.
    year : integer
        Year
    csvfile: boolean, optional
        Set to True if you prefer csv to plain txt.
        default is False.
    plt : boolean, optional
        To print plots to screen or not.
        default is TRUE.
    extension : string, optional
        Solution subdirectory.
        default is empty string.
    doy1 : integer, optional
        Initial day of year
        default is 1.
    doy2 : integer, optional
        End day of year.
        default is 366.
    ampl: float 
        New amplitude constraint
        default is 0.
    azim1: int, optional
        New min azimuth
        default is 0.
    azim2: int, optional
        New max azimuth
        default is 360.
    h1: float optional 
        lowest allowed reflector height
        default is 0
    h2: float optional 
        highest allowed reflector height
        default is 300
    peak2noise: float, optional
        New peak to noise constraint
        default is 0.

    """

    if len(station) != 4:
        print('station names must be four characters long')
        sys.exit()

    # make surer environment variables are set
    g.check_environ_variables()
    xdir = os.environ['REFL_CODE']

    # set subdirectory for results
    subdir = station
    g.set_subdir(subdir)

    txtdir = xdir + '/Files/' + subdir
    print(txtdir)

    #does not currently allow this - but should
    writecsv = False

    if (h1 > h2):
        print('h1 must be less than h2. You submitted ', h1, ' and ', h2)
        sys.exit()

    # these are things not needed by rh_plot - but the code expects it
    sigma = 3; txtfile = ''; kplt=False
    default_usage = False
    hires_figs = False; 
    fs = 12
    ntv, obstimes, fname, fname_new = t.readin_and_plot(station, year, doy1, doy2, plt, extension, sigma,
            writecsv, azim1, azim2, ampl, peak2noise, txtfile,h1,h2,kplt,txtdir,default_usage, hires_figs,fs)

    if plt:
        myplt.show()

def main():
    args = parse_arguments()
    rh_plot(**args)


if __name__ == "__main__":
    main()



import numpy as np
import re
import datetime
import struct

import gnssrefl.gps as g

# Joakim Strandberg wrote this code originally-
# I made some small changes ???  Kristine M. Larson

class RinexError(Exception):
    pass


def getrinexversion(filename):
    """ Scan the file for RINEX version number.

    Parameters
    ----------
    filename : str
        Filename of the rinex file

    Returns
    -------
    version : str
        Version number.
    """
    with open(filename, 'r') as f:
        line = f.readline()
        while 'RINEX VERSION / TYPE' not in line:
            line = f.readline()

            if 'END OF HEADER' in line:
                raise RinexError("No 'RINEX VERSION / TYPE' found.")
    return line[:9].strip()


def readheader(lines, rinexversion):
    # See no reason for keeping this public. Should have been private from the start.
    print("WARNING! Deprecated!")
    return _readheader(lines, rinexversion)


def _readheader(lines, rinexversion):
    """ Read and return header information for the RINEX file

    Parameters
    ----------
    lines : list[str]
        List of each line in the RINEX file.

    rinexversion : str
        Version number for the RINEX file

    Returns
    -------
    header : dict
        Dict containing the header information from the RINEX file.

    headerlines : list[int]
        List of starting line for the headers of each data block.

    headerlengths : list[int]
        List of length for the headers of each data block. 'None' for rinex3.

    obstimes : list[datetime.datetime]
        List of time of measurement for each measurement epoch.

    satlists : list[list[str]]
        List containing lists of satellites present in each block.

    satset : set(str)
        Set containing all satellites in the data.
    """
    try:
        if '2.1' in rinexversion:
            return _readheader_v21x(lines)
        elif '3' in rinexversion:
            return _readheader_v3(lines)
        else:
            raise RinexError('RINEX v%s is not supported.' % rinexversion)

    except KeyError as e:
        raise RinexError('Missing required header %s' % str(e))


def _readheader_v21x(lines):
    """ Read rinex version 2.10 and 2.11 
    """

    header = {}
    # Capture header info

    for i, line in enumerate(lines):
        if "END OF HEADER" in line:
            i += 1  # skip to data
            break

        if line[60:80].strip() not in header:  # Header label
            header[line[60:80].strip()] = line[:60]  # don't strip for fixed-width parsers
            # string with info
        else:
            header[line[60:80].strip()] += "\n"+line[:60]
            # concatenate to the existing string

    rowpersat = 1 + (len(header['# / TYPES OF OBSERV'][6:].split())-1) // 5

    timeoffirstobs = [part for part in header['TIME OF FIRST OBS'].split()]

    headerlines = []
    headerlengths = []
    obstimes = []
    epochsatlists = []
    # for those of who do not like datetime
    gpstime = np.empty(shape=[0, 2])
    satset = set()

    century = int(timeoffirstobs[0][:2]+'00')
    # This will result in an error if the record overlaps the end of the century. So if someone feels this is a major
    # problem, feel free to fix it. Personally can't bother to do it...

    pattern = re.compile('(\s{2}\d|\s\d{2}){2}')

    while i < len(lines):
        if pattern.match(lines[i][:6]):  # then it's the first line in a header record
            if int(lines[i][28]) in (0, 1, 6):  # CHECK EPOCH FLAG  STATUS
                headerlines.append(i)
                year, month, day, hour = lines[i][1:3], lines[i][4:6], lines[i][7:9], lines[i][10:12]
                minute, second = lines[i][13:15], lines[i][16:26]
                obstimes.append(datetime.datetime(year=century+int(year),
                                                  month=int(month),
                                                  day=int(day),
                                                  hour=int(hour),
                                                  minute=int(minute),
                                                  second=int(float(second)),
                                                  microsecond=int(float(second) % 1 * 100000)))

                week, sow = g.kgpsweek(century+int(year), int(month), int(day), int(hour), int(minute), int(float(second)))
                ne = np.array((week,sow))
                gpstime = np.vstack((gpstime,ne))

                numsats = int(lines[i][29:32])  # Number of visible satellites %i3
                headerlengths.append(1 + (numsats-1)//12)  # number of lines in header, depends on how many svs on view

                if numsats > 12:
                    sv = []
                    for s in range(numsats):
                        if s > 0 and s % 12 == 0:
                            i += 1
                        sv.append(lines[i][32+(s % 12)*3:35+(s % 12)*3])
                    epochsatlists.append(sv)

                else:
                    epochsatlists.append([lines[i][32+s*3:35+s*3] for s in range(numsats)])

                i += numsats*rowpersat+1

            else:  # there was a comment or some header info
                flag = int(lines[i][28])
                if flag != 4:
                    print(flag)
                skip = int(lines[i][30:32])
                i += skip+1
        else:
            # We have screwed something up and have to iterate to get to the next header row, or eventually the end.
            i += 1

    for satlist in epochsatlists:
        satset = satset.union(satlist)

    return header, headerlines, headerlengths, obstimes, epochsatlists, satset, gpstime 


def _readheader_v3(lines):
    """ Read rinex version 3 """

    header = {}
    # Capture header info

    for i, line in enumerate(lines):
        if "END OF HEADER" in line:
            i += 1  # skip to data
            break

        if line[60:80].strip() not in header:  # Header label
            header[line[60:80].strip()] = line[:60]  # don't strip for fixed-width parsers
            # string with info
        else:
            header[line[60:80].strip()] += "\n"+line[:60]
            # concatenate to the existing string

    headerlines = []
    obstimes = []
    epochsatlists = []
    satset = set()

    while i < len(lines):
        if lines[i][0] == '>':  # then it's the first line in a header record
            if int(lines[i][31]) in (0, 1, 6):  # CHECK EPOCH FLAG  STATUS
                headerlines.append(i)
                year, month, day, hour = lines[i][2:6], lines[i][7:9], lines[i][10:12], lines[i][13:15]
                minute, second = lines[i][16:18], lines[i][19:30]
                obstimes.append(datetime.datetime(year=int(year),
                                                  month=int(month),
                                                  day=int(day),
                                                  hour=int(hour),
                                                  minute=int(minute),
                                                  second=int(float(second)),
                                                  microsecond=int(float(second) % 1 * 100000)))

                numsats = int(lines[i][33:35])  # Number of visible satellites %i3

                sv = []
                for j in range(numsats):
                    sv.append(lines[i+1+j][:3])

                i += numsats+1
                epochsatlists.append(sv)

            else:  # there was a comment or some header info
                flag = int(lines[i][31])
                if flag != 4:
                    print(flag)
                skip = int(lines[i][30:32])
                i += skip+1
        else:
            # We have screwed something up and have to iterate to get to the next header row, or eventually the end.
            i += 1

    for satlist in epochsatlists:
        satset = satset.union(satlist)

    headerlengths = None
    return header, headerlines, headerlengths, obstimes, epochsatlists, satset


def _converttofloat(numberstr):
    try:
        return float(numberstr)
    except ValueError:
        return np.nan


def _readblocks(lines, rinexversion, header, headerlines, headerlengths, epochsatlists, satset):
    """ Read and return information in the blocks for the RINEX file

    Parameters
    ----------
    lines : list[str]
        List of each line in the RINEX file.

    rinexversion : str
        Version number for the RINEX file.

    header : dict
        Dict containing the header information from the RINEX file.

    headerlines : list[int]
        List of starting line for the headers of each data block.

    headerlengths : list[int]
        List of length for the headers of each data block.

    satlists : list[list[str]]
        List containing lists of satellites present in each block.

    satset : set(str)
        Set containing all satellites in the data.

    Returns
    -------
    observationdata : dict
        Dict with data-arrays.

    satlists : dict
        Dict with lists of visible satellites.

    prntoidx : dict
        Dict with translation dicts.

    obstypes : dict
        Dict with observation types.
    """
    try:
        if '2.1' in rinexversion:
            return _readblocks_v21(lines, header, headerlines, headerlengths, epochsatlists, satset)
        elif '3' in rinexversion:
            return _readblocks_v3(lines, header, headerlines, epochsatlists, satset)
        else:
            raise RinexError('RINEX v%s is not supported.' % rinexversion)

    except KeyError as e:
        raise RinexError('Missing required header %s' % str(e))



def _readblocks_v21(lines, header, headerlines, headerlengths, epochsatlists, satset):
    """ Read the lines of data.

    Parameters
    ----------
    lines : list[str]
        List of each line in the RINEX file.

    header : dict
        Dict containing the header information from the RINEX file.

    headerlines : list[int]
        List of starting line for the headers of each data block.

    headerlengths : list[int]
        List of length for the headers of each data block.

    satlists : list[list[str]]
        List containing lists of satellites present in each block.

    satset : set(str)
        Set containing all satellites in the data.

    Returns
    -------
    observationdata : dict
        Dict with data-arrays.

    satlists : dict
        Dict with lists of visible satellites.

    prntoidx : dict
        Dict with translation dicts.

    obstypes : dict
        Dict with observation types.

    See also
    --------
    processrinexfile : The wrapper.
    """
    observables = header['# / TYPES OF OBSERV'][6:].split()
    nobstypes = len(observables)
    rowpersat = 1 + (nobstypes-1) // 5
    nepochs = len(headerlines)

    systemletters = set([letter for letter in set(''.join(satset)) if letter.isalpha()])
    satlists = {letter: [] for letter in systemletters}

    observationdata = {}
    prntoidx = {}
    obstypes = {}

    for sat in satset:
        satlists[sat[0]].append(int(sat[1:]))

    for letter in systemletters:
        satlists[letter].sort()
        nsats = len(satlists[letter])
        observationdata[letter] = np.nan * np.zeros((nepochs, nsats, nobstypes))
        prntoidx[letter] = {prn: idx for idx, prn in enumerate(satlists[letter])}
        obstypes[letter] = observables  # Proofing for V3 functionality

    fmt = '14s 2x '*nobstypes
    fieldstruct = struct.Struct(fmt)
    parse = fieldstruct.unpack_from

    for iepoch, (headerstart, headerlength, satlist) in enumerate(zip(headerlines, headerlengths, epochsatlists)):
        for i, sat in enumerate(satlist):
            datastring = ''.join(["{:<80}".format(line.rstrip()) for line in
                                  lines[headerstart+headerlength+rowpersat*i:headerstart+headerlength+rowpersat*(i+1)]])
            try:
                data = np.array([_converttofloat(number.decode('ascii')) for number in parse(datastring.encode('ascii'))])

                systemletter = sat[0]
                prn = int(sat[1:])

                observationdata[systemletter][iepoch, prntoidx[systemletter][prn], :] = data
            except struct.error:
                continue

    for letter in observationdata:
        kept_observables = [i for i in range(len(obstypes[letter])) if np.sum(~np.isnan(observationdata[letter][:,:,i]))>0]
        observationdata[letter] = observationdata[letter][:, :, kept_observables]
        obstypes[letter] = [obstypes[letter][i] for i in kept_observables]

    return observationdata, satlists, prntoidx, obstypes


def _readblocks_v3(lines, header, headerlines, epochsatlists, satset):
    """ Read the lines of data for rinex 3 files.

    Parameters
    ----------
    lines : list[str]
        List of each line in the RINEX file.

    header : dict
        Dict containing the header information from the RINEX file.

    headerlines : list[int]
        List of starting line for the headers of each data block.

    satlists : list[list[str]]
        List containing lists of satellites present in each block.

    satset : set(str)
        Set containing all satellites in the data.

    Returns
    -------
    observationdata : dict
        Dict with data-arrays.

    satlists : dict
        Dict with lists of visible satellites.

    prntoidx : dict
        Dict with translation dicts.

    obstypes : dict
        Dict with observation types.

    See also
    --------
    processrinexfile : The wrapper.
    """
    nepochs = len(headerlines)

    obstypes = {}
    systemletter = ''
    for line in header['SYS / # / OBS TYPES'].splitlines():
        if line[0] != ' ':
            systemletter = line[0]
            obstypes[systemletter] = line[6:].split()
        else:
            # It's a continuation line and we just add the obstypes
            obstypes[systemletter].extend(line[6:].split())

    systemletters = [key for key in obstypes]
    satlists = {letter: [] for letter in systemletters}

    observationdata = {}
    prntoidx = {}
    parser = {}

    for sat in satset:
        satlists[sat[0]].append(int(sat[1:]))

    for letter in systemletters:

        if len(satlists[letter]) == 0:
            # No data exist for the system. Delete and skip.
            satlists.pop(letter)
            continue

        satlists[letter].sort()
        nsats = len(satlists[letter])
        nobstypes = len(obstypes[letter])
        observationdata[letter] = np.nan * np.zeros((nepochs, nsats, nobstypes))
        prntoidx[letter] = {prn: idx for idx, prn in enumerate(satlists[letter])}

        fmt = '3x' + '14s 2x ' * (nobstypes-1) + '14s'
        parser[letter] = struct.Struct(fmt)

    for iepoch, (headerstart, satlist) in enumerate(zip(headerlines, epochsatlists)):
        for i, sat in enumerate(satlist):
            datastring = lines[headerstart+1+i]

            systemletter = sat[0]
            prn = int(sat[1:])

            try:
                data = np.array([_converttofloat(number.decode('ascii'))
                                 for number in parser[systemletter].unpack_from(datastring.encode('ascii'))])

                observationdata[systemletter][iepoch, prntoidx[systemletter][prn], :] = data
            except struct.error:
                continue
                #print('Data loss for %s at epoch %i' % (sat, iepoch))

    for letter in observationdata:
        kept_observables = [i for i in range(len(obstypes[letter])) if np.sum(~np.isnan(observationdata[letter][:,:,i]))>0]
        observationdata[letter] = observationdata[letter][:, :, kept_observables]
        obstypes[letter] = [obstypes[letter][i] for i in kept_observables]

    return observationdata, satlists, prntoidx, obstypes


def processrinexfile(filename, savefile=None):
    """ Process a RINEX file into python format

    Parameters
    ----------
    filename : str
        Filename of the rinex file

    savefile : str, optional
        Name of file to save data to. If supplied the data is saved to a compressed npz file.

    Returns
    -------
    observationdata : dict
        Dict with a nobs x nsats x nobstypes nd-array for each satellite constellation containing the measurements. The
        keys of the dict correspond to the systemletter as used in RINEX files (G for GPS, R for GLONASS, etc).

        nobs is the number of observations in the RINEX data, nsats the number of visible satellites for the particular
        system during the whole measurement period, and nobstypes is the number of different properties recorded.

    satlists : dict
        Dict containing the full list of visible satellites during the whole measurement period for each satellite
        constellation.

    prntoidx : dict
        Dict which for each constellation contains a dict which translates the PRN number into the index of the
        satellite in the observationdata array.

    obstypes : dict
        Dict containing the observables recorded for each satellite constellation.

    header : dict
        Dict containing the header information from the RINEX file.

    obstimes : list[datetime.datetime]
        List of time of measurement for each measurement epoch.

    """
    rinexversion = getrinexversion(filename)

    with open(filename, 'r') as f:
        lines = f.read().splitlines(True)

    header, headerlines, headerlengths, obstimes, epochsatlists, satset,gpstime = _readheader(lines, rinexversion)
    observationdata, satlists, prntoidx, obstypes = _readblocks(lines, rinexversion, header, headerlines,
                                                                headerlengths, epochsatlists, satset)

    if savefile is not None:
        saverinextonpz(savefile, observationdata, satlists, prntoidx, obstypes, header, obstimes)

    return observationdata, satlists, prntoidx, obstypes, header, obstimes, gpstime 


def mergerinexfiles(filelist, savefile=None):
    """ Process several rinexfiles and merges them into one file.

    Can be used to for example merge several rinexfiles from the same day to a single file. All files must be from the
    same receiver and have the same version. No guarantees are given if files are from different receivers.

    !!! Currently only functional for RINEX3. !!!


    Parameters
    ----------
    filename : str
        Filename of the rinex file

    savefile : str, optional
        Name of file to save data to. If supplied the data is saved to a compressed npz file.

    Returns
    -------
    observationdata : dict
        Dict with a nobs x nsats x nobstypes nd-array for each satellite constellation containing the measurements. The
        keys of the dict correspond to the systemletter as used in RINEX files (G for GPS, R for GLONASS, etc).

        nobs is the number of observations in the RINEX data, nsats the number of visible satellites for the particular
        system during the whole measurement period, and nobstypes is the number of different properties recorded.

    satlists : dict
        Dict containing the full list of visible satellites during the whole measurement period for each satellite
        constellation.

    prntoidx : dict
        Dict which for each constellation contains a dict which translates the PRN number into the index of the
        satellite in the observationdata array.

    obstypes : dict
        Dict containing the observables recorded for each satellite constellation.

    header : dict
        Dict containing the header information from the first RINEX file.

    obstimes : list[datetime.datetime]
        List of time of measurement for each measurement epoch.
    """

    rinexversion = getrinexversion(filelist[0])

    if '3' in rinexversion:

        with open(filelist[0], 'r') as f:
            lines = f.read().splitlines(True)

        if len(filelist) > 1:
            for filename in filelist[1:]:
                with open(filename, 'r') as f:
                    line = f.readline()
                    while 'END OF HEADER' not in line:
                        line = f.readline()
                    lines.extend(f.read().splitlines(True))

        header, headerlines, headerlengths, obstimes, epochsatlists, satset = _readheader(lines, rinexversion)
        observationdata, satlists, prntoidx, obstypes = _readblocks(lines, rinexversion, header, headerlines,
                                                                    headerlengths, epochsatlists, satset)

        if savefile is not None:
            saverinextonpz(savefile, observationdata, satlists, prntoidx, obstypes, header, obstimes)

        return observationdata, satlists, prntoidx, obstypes, header, obstimes



    else:
        NotImplementedError('Multiple file merging is not implemented for RINEX 2 yet.')
        #TODO: implement

        #for filename in filelist:
        #    if not getrinexversion(filename) == rinexversion:
        #        raise RinexError('All files are not the same RINEX version.')

        #observationdata, satlists, prntoidx, obstypes, header, obstimes = processrinexfile(filename[0], savefile=None)

        #if len(filelist) > 1:
        #    for filename in filelist[1:]:
        #        observationdata_add, satlists_add, prntoidx_add, obstypes_add, header_add, obstimes_add = processrinexfile(filename, savefile=None)

        #        for key in observationdata:
        #            if satlists[key]
        #                observationdata[key] = np.append(observationdata[key], observationdata_add[key])

        # On another level? Sätta ihop data tidigare, dvs läsa alla headers först?



def separateobservables(observationdata, obstypes):
    """
    Parameters
    ----------
    observationdata : dict
        Data dict as returned by processrinexfile, or loadrinexfromnpz.

    obstypes : dict
        Dict with observation types for each system as returned by processrinexfile, or loadrinexfromnpz.

    Returns
    -------
    separatedobservationdata : dict
        Dict for each system where the data for each observable is separated into its own dict. I.e. to access the P1
        data for GPS from a RINEX2 file it is only necessary to write `separatedobservationdata['G']['C1']`.
    """

    separatedobservationdata = dict()
    for systemletter in observationdata:
        separatedobservationdata[systemletter] = dict()
        for idx, obstype in enumerate(obstypes[systemletter]):
            separatedobservationdata[systemletter][obstype] = observationdata[systemletter][:, :, idx]
    return separatedobservationdata


def saverinextonpz(savefile, observationdata, satlists, prntoidx, obstypes, header, obstimes):
    """ Save data to numpy's npz format.

    Parameters
    ----------
    savefile : str
        Path to where to save the data.

    observationdata, satlists, prntoidx, obstypes, header, obstimes: dict
        Data as returned from processrinexfile

    See Also
    --------
    processrinexfile
    """
    savestruct = {'systems': []}

    for systemletter in observationdata:
        savestruct[systemletter+'systemdata'] = observationdata[systemletter]
        savestruct[systemletter+'systemsatlists'] = satlists[systemletter]
        savestruct[systemletter+'prntoidx'] = prntoidx[systemletter]
        savestruct[systemletter+'obstypes'] = obstypes[systemletter]
        savestruct['systems'].append(systemletter)

    savestruct['obstimes'] = obstimes
    savestruct['header'] = header

    np.savez_compressed(savefile, **savestruct)


def loadrinexfromnpz(npzfile):
    """ Load data previously stored in npz-format

    Parameters
    ----------
    npzfile : str
        Path to the stored data.

    Returns
    -------
    observationdata, satlists, prntoidx, obstypes, header, obstimes: dict
        Data in the same format as returned by processrinexfile
    """
    rawdata = np.load(npzfile)

    observationdata = {}
    satlists = {}
    prntoidx = {}
    obstypes = {}

    for systemletter in rawdata['systems']:
        observationdata[systemletter] = rawdata[systemletter+'systemdata']
        satlists[systemletter] = list(rawdata[systemletter+'systemsatlists'])
        prntoidx[systemletter] = rawdata[systemletter+'prntoidx'].item()
        obstypes[systemletter] = list(rawdata[systemletter+'obstypes'])

    header = rawdata['header'].item()
    obstimes = list(rawdata['obstimes'])

    return observationdata, satlists, prntoidx, obstypes, header, obstimes
